from __future__ import annotations

import asyncio
import itertools
from collections.abc import Callable
from contextvars import ContextVar, Token
from types import TracebackType
from typing import Self

from .types import TaskGroupError

__all__ = (
    "TaskGroup",
    "TaskGroupError",
    "current_taskgroup",
)

current_taskgroup: ContextVar[TaskGroup] = ContextVar("current_taskgroup")


class TaskGroup(asyncio.TaskGroup):
    _name: str
    _current_taskgroup_token: Token[TaskGroup]

    def __init__(self, *, name: str | None = None) -> None:
        super().__init__()
        if name is None:
            self._name = f"tg-{_name_counter()}"
        else:
            self._name = str(name)

    def get_name(self) -> str:
        return self._name

    async def __aenter__(self) -> Self:
        self._current_taskgroup_token = current_taskgroup.set(self)
        return await super().__aenter__()

    async def __aexit__(
        self,
        et: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        try:
            await super().__aexit__(et, exc, tb)
        except BaseExceptionGroup as eg:
            # Just wrap the exception group as TaskGroupError for backward
            # compatibility.  In Python 3.11 or higher, TaskGroupError
            # also inherits BaseExceptionGroup, so the standard except*
            # clauses and ExceptionGroup methods will work as expected.
            # New codes should migrate to directly using asyncio.TaskGroup.
            raise TaskGroupError(eg.message, eg.exceptions) from None
        finally:
            current_taskgroup.reset(self._current_taskgroup_token)


_name_counter: Callable[[], int] = itertools.count(1).__next__



"""
Provides a simple implementation of timers run inside asyncio event loops.
"""

import asyncio
import contextlib
import enum
import functools
from collections.abc import Callable, Iterator
from typing import TypeVar
from unittest import mock

from .cancel import cancel_and_wait
from .compat import get_running_loop
from .taskscope import TaskScope
from .types import CoroutineLike

__all__ = (
    "create_timer",
    "TimerDelayPolicy",
    "VirtualClock",
)

T_SelectorRet = TypeVar("T_SelectorRet")


class TimerDelayPolicy(enum.Enum):
    """
    An enumeration of supported policies for when the timer function takes
    longer on each tick than the given timer interval.
    """

    DEFAULT = 0
    CANCEL = 1


def create_timer(
    cb: Callable[[float], CoroutineLike[None]],
    interval: float,
    delay_policy: TimerDelayPolicy = TimerDelayPolicy.DEFAULT,
    loop: asyncio.AbstractEventLoop | None = None,
) -> asyncio.Task[None]:
    """
    Schedule a timer with the given callable and the interval in seconds.
    The interval value is also passed to the callable.
    If the callable takes longer than the timer interval, all accumulated
    callable's tasks will be cancelled when the timer is cancelled.

    Args:
        cb: An async timer callback function accepting the configured interval value.

    Returns:
        You can stop the timer by cancelling the returned task.

    .. versionchanged:: 2.0

       Unhandled exceptions in the tick tasks no longer terminates the timer task,
       as it is changed to use :class:`~aiotools.taskscope.TaskScope` instead of
       :class:`~aiotools.taskgroup.TaskGroup`.

    .. versionchanged:: 2.0

       Cancelling and awaiting the task returned by ``create_timer()`` now raises
       :exc:`asyncio.CancelledError`.  Use :func:`~aiotools.cancel.cancel_and_wait()`
       to safely consume or re-raise it.
    """

    async def _timer() -> None:
        fired_tasks: list[asyncio.Task[None]] = []
        async with TaskScope() as ts:
            while True:
                if delay_policy == TimerDelayPolicy.CANCEL:
                    await asyncio.gather(
                        *(cancel_and_wait(t) for t in fired_tasks),
                        return_exceptions=True,
                    )
                    fired_tasks.clear()
                else:
                    fired_tasks[:] = [t for t in fired_tasks if not t.done()]
                t = ts.create_task(cb(interval))
                fired_tasks.append(t)
                await asyncio.sleep(interval)

    return asyncio.create_task(_timer())


class VirtualClock:
    """
    Provide a virtual clock for an asyncio event loop
    which makes timing-based tests deterministic and instantly completed.
    """

    def __init__(self) -> None:
        self.vtime = 0.0

    def virtual_time(self) -> float:
        """
        Return the current virtual time.
        """
        return self.vtime

    def _virtual_select(
        self,
        orig_select: Callable[[float | None], T_SelectorRet],
        timeout: float | None,
    ) -> T_SelectorRet:
        if timeout is not None:
            self.vtime += timeout
        return orig_select(0)  # override the timeout to zero

    @contextlib.contextmanager
    def patch_loop(self) -> Iterator[None]:
        """
        Override some methods of the current event loop
        so that sleep instantly returns while proceeding the virtual clock.
        """
        loop = get_running_loop()
        # Both SelectorEventLoop and ProactorEventLoop has "_selector" with "select()"
        # method which accepts a timeout value until the next wake-up point.
        assert hasattr(loop, "_selector"), (
            "Only stdlib's SelectorEventLoop or ProactorEventLoop is supported."
        )
        with (
            mock.patch.object(
                loop._selector,
                "select",
                new=functools.partial(self._virtual_select, loop._selector.select),
            ),
            mock.patch.object(
                loop,
                "_clock_resolution",
                new=0.001,
            ),
            mock.patch.object(
                loop,
                "time",
                new=self.virtual_time,
            ),
        ):
            yield



"""
This module implements a simple :func:`os.fork()`-like interface,
but in an asynchronous way with full support for PID file descriptors
on Python 3.9 or higher and the Linux kernel 5.4 or higher.

It internally synchronizes the beginning and readiness status of child processes
so that the users may assume that the child process is completely interruptible after
:func:`afork()` returns.

.. versionchanged:: 1.9.0

    The internal implementation has changed to use :py:mod:`multiprocessing` to embrace
    ``posix_spawn()`` in macOS and Windows platforms.
    This introduces a potential BREAKING CHANGE that users now must pass module-level
    functions or class methods as the target function of :func:`afork()`, so that they
    can be imported in the new subprocess.
"""

from __future__ import annotations

import asyncio
import errno
import logging
import multiprocessing as mp
import multiprocessing.connection as mpc
import multiprocessing.context as mpctx
import os
import signal
import sys
import traceback
from abc import ABCMeta, abstractmethod
from collections.abc import Callable
from typing import ClassVar, TypeAlias

from .compat import get_running_loop

__all__ = (
    "AbstractChildProcess",
    "PosixChildProcess",
    "PidfdChildProcess",
    "afork",
)

logger = logging.getLogger(__name__)

if sys.platform != "win32":
    MPProcess: TypeAlias = (
        mpctx.Process | mpctx.SpawnProcess | mpctx.ForkProcess | mpctx.ForkServerProcess
    )
    MPContext: TypeAlias = (
        mpctx.DefaultContext
        | mpctx.ForkContext
        | mpctx.ForkServerContext
        | mpctx.SpawnContext
    )

else:
    MPProcess: TypeAlias = mpctx.Process | mpctx.SpawnProcess
    MPContext: TypeAlias = mpctx.DefaultContext | mpctx.SpawnContext

_has_pidfd = False
if hasattr(os, "pidfd_open"):
    # signal.pidfd_send_signal() is available in Linux kernel 5.1+
    # and os.pidfd_open() is available in Linux kernel 5.3+.
    # So let's check with os.pidfd_open() which requires higher version.
    try:
        os.pidfd_open(0, 0)  # type: ignore[attr-defined]
    except OSError as e:
        if e.errno in (errno.EBADF, errno.EINVAL):
            _has_pidfd = True
        # if the kernel does not support this,
        # it will say errno.ENOSYS or errno.EPERM


class AbstractChildProcess(metaclass=ABCMeta):
    """
    The abstract interface to control and monitor a forked child process.
    """

    @property
    @abstractmethod
    def pid(self) -> int:
        """
        The process ID of the child process.
        """
        raise NotImplementedError

    @abstractmethod
    def send_signal(self, signum: int) -> None:
        """
        Send a UNIX signal to the child process.
        If the child process is already terminated, it will log
        a warning message and return.
        """
        raise NotImplementedError

    @abstractmethod
    async def wait(self) -> int:
        """
        Wait until the child process terminates or reclaim the child process' exit
        code if already terminated.
        If there are other coroutines that has waited the same process, it may
        return 255 and log a warning message.
        """
        raise NotImplementedError


class PosixChildProcess(AbstractChildProcess):
    """
    A POSIX-compatible version of :class:`AbstractChildProcess`.

    .. versionchanged:: 1.9.0

       The :meth:`wait()` method now polls the status of child process instead of
       launching a separate thread that makes a blocking ``os.waitpid()`` syscall.
       The polling interval may be adjusted using the class attribute
       :attr:`poll_interval`, which defaults to 50 msec.
    """

    poll_interval: ClassVar[float] = 0.05

    def __init__(self, proc: MPProcess, pid: int) -> None:
        self._proc = proc
        self._pid = pid
        self._terminated = False
        self._returncode: int | None = None

    @property
    def pid(self) -> int:
        return self._pid

    def send_signal(self, signum: int) -> None:
        if self._terminated:
            if signum != signal.SIGKILL:
                logger.warning(
                    "PosixChildProcess(%d).send_signal(%d): "
                    "The process has already terminated.",
                    self._pid,
                    signum,
                )
            return
        if signum == signal.SIGKILL:
            logger.warning("Force-killed hanging child: %d", self._pid)
        try:
            os.kill(self._pid, signum)
        except ProcessLookupError:
            logger.warning(
                "PosixChildProcess(%d).send_signal(%d): "
                "The process has already terminated.",
                self._pid,
                signum,
            )

    async def wait(self) -> int:
        status = 0
        while self._returncode is None:
            try:
                while True:
                    pid, status = os.waitpid(self._pid, os.WNOHANG)
                    if pid == self._pid:
                        self._returncode = os.waitstatus_to_exitcode(status)
                        break
                    await asyncio.sleep(self.poll_interval)
            except ChildProcessError:
                # Since we have already checked for the process's existence,
                # we can assume that the process has terminated.
                self._returncode = 255
        self._proc.join()  # let multiprocessing clean up itself
        self._terminated = True
        return self._returncode


class PidfdChildProcess(AbstractChildProcess):
    """
    A pidfd (PID file descriptor) based version of :class:`AbstractChildProcess`.

    The main advantage of pidfd is that we no longer need to actively poll
    the child status, making the whole operation async-native.

    Note that pidfd may not be available on all Linux systems
    depending on at which Linux kernel version your Python executable is built.
    """

    def __init__(self, proc: MPProcess, pid: int, pidfd: int) -> None:
        self._proc = proc
        self._pid = pid
        self._pidfd = pidfd
        self._returncode: int | None = None
        self._wait_event = asyncio.Event()
        self._terminated = False
        loop = get_running_loop()
        loop.add_reader(self._pidfd, self._do_wait)

    @property
    def pid(self) -> int:
        return self._pid

    def send_signal(self, signum: int) -> None:
        if self._terminated:
            if signum != signal.SIGKILL:
                logger.warning(
                    "PidfdChildProcess(%d, %d).send_signal(%d): "
                    "The process has already terminated.",
                    self._pid,
                    self._pidfd,
                    signum,
                )
            return
        if signum == signal.SIGKILL:
            logger.warning("Force-killed hanging child: %d", self._pid)
        signal.pidfd_send_signal(self._pidfd, signum)  # type: ignore

    def _do_wait(self) -> None:
        loop = get_running_loop()
        try:
            # The flag is WEXITED | __WALL from linux/wait.h
            # (__WCLONE value is out of range of int...)
            status_info = os.waitid(
                os.P_PIDFD,
                self._pidfd,
                os.WEXITED | 0x40000000,
            )
        except ChildProcessError:
            # The child process is already reaped
            # (may happen if waitpid() is called elsewhere).
            self._returncode = 255
            logger.warning(
                "child process %d exit status already read: "
                "it will report returncode 255",
                self._pid,
            )
        else:
            assert (
                status_info is not None
            )  # Always available since we didn't set WNOHANG
            if status_info.si_code == os.CLD_KILLED:
                self._returncode = -status_info.si_status  # signal number
            elif status_info.si_code == os.CLD_EXITED:
                self._returncode = status_info.si_status
            elif status_info.si_code == os.CLD_DUMPED:
                self._returncode = -status_info.si_status  # signal number
            else:
                logger.warning(
                    "unexpected si_code %d and si_status %d for child process %d",
                    status_info.si_code,
                    status_info.si_status,
                    self._pid,
                )
                self._returncode = 255
        finally:
            loop.remove_reader(self._pidfd)
            os.close(self._pidfd)
            self._proc.join()  # let multiprocessing clean up itself
            self._terminated = True
            self._wait_event.set()

    async def wait(self) -> int:
        await self._wait_event.wait()
        assert self._returncode is not None
        return self._returncode


def _child_main(
    write_pipe: mpc.Connection,
    child_func: Callable[[], int],
) -> int:
    ret = -255
    # Reset signal handlers to default for proper signal handling.
    # Multiprocessing may set SIGINT to SIG_IGN to prevent KeyboardInterrupt
    # in worker processes, but we want the default behavior where SIGINT
    # raises KeyboardInterrupt so that child_func can handle it properly.
    signal.signal(signal.SIGINT, signal.default_int_handler)
    signal.signal(signal.SIGTERM, signal.SIG_DFL)
    try:
        # notify the parent that the child is ready to execute the requested function.
        write_pipe.send_bytes(b"\0")
        write_pipe.close()
        ret = child_func()
    except KeyboardInterrupt:
        ret = -signal.SIGINT
    except SystemExit:
        ret = -signal.SIGTERM
    except Exception:
        traceback.print_exc()
    finally:
        os._exit(ret)
    return ret


async def _fork_posix(
    child_func: Callable[[], int],
    mp_context: MPContext,
) -> tuple[MPProcess, int]:
    loop = get_running_loop()
    read_pipe, write_pipe = mp_context.Pipe()
    proc = mp_context.Process(
        target=_child_main,
        args=(write_pipe, child_func),
        daemon=True,
    )
    proc.start()
    assert proc.pid is not None
    pid = proc.pid

    # Wait for the child's readiness notification
    init_event = asyncio.Event()
    loop.add_reader(read_pipe.fileno(), init_event.set)
    await init_event.wait()
    loop.remove_reader(read_pipe.fileno())
    read_pipe.recv_bytes(1)
    read_pipe.close()
    return proc, pid


async def _clone_pidfd(
    child_func: Callable[[], int],
    mp_context: MPContext,
) -> tuple[MPProcess, int, int]:
    loop = get_running_loop()
    read_pipe, write_pipe = mp_context.Pipe()
    proc = mp_context.Process(
        target=_child_main,
        args=(write_pipe, child_func),
        daemon=True,
    )
    proc.start()
    assert proc.pid is not None
    pid = proc.pid
    fd = os.pidfd_open(pid, 0)  # type: ignore

    # Wait for the child's readiness notification
    init_event = asyncio.Event()
    loop.add_reader(read_pipe.fileno(), init_event.set)
    await init_event.wait()
    loop.remove_reader(read_pipe.fileno())
    read_pipe.recv_bytes(1)
    read_pipe.close()
    return proc, pid, fd


async def afork(
    child_func: Callable[[], int],
    *,
    mp_context: MPContext | None = None,
) -> AbstractChildProcess:
    """
    Fork the current process and execute the given function in the child.
    The return value of the function will become the exit code of the child
    process.

    Args:
        child_func: A function that represents the main function of the child and
                    returns an integer as its exit code.
                    Note that the function must set up a new event loop if it
                    wants to run asyncio codes.
        mp_context: The multiprocessing context to use. If not provided, the default
                    context will be used.

    .. versionadded:: 1.9.0

        The argument ``mp_context``.
    """
    if mp_context is None:
        mp_context = mp.get_context()
    if _has_pidfd:
        proc, pid, pidfd = await _clone_pidfd(child_func, mp_context)
        return PidfdChildProcess(proc, pid, pidfd)
    else:
        proc, pid = await _fork_posix(child_func, mp_context)
        return PosixChildProcess(proc, pid)



from __future__ import annotations

import asyncio
import itertools
import logging
import sys
import traceback
import weakref
from collections.abc import Awaitable, Callable, Sequence
from contextvars import ContextVar, Token
from types import TracebackType
from typing import (
    Any,
    Self,
    TypeVar,
    cast,
)

from typing_extensions import deprecated

from aiotools.types import CoroutineLike

from .. import compat
from .types import AsyncExceptionHandler

__all__ = (
    "PersistentTaskGroup",
    "current_ptaskgroup",
)

current_ptaskgroup: ContextVar[PersistentTaskGroup] = ContextVar("current_ptaskgroup")  # type: ignore[deprecated]

_ptaskgroup_idx = itertools.count()
_log = logging.getLogger(__name__)
_all_ptaskgroups: weakref.WeakSet[PersistentTaskGroup] = weakref.WeakSet()  # type: ignore[deprecated]

T = TypeVar("T")


async def _default_exc_handler(
    exc_type: type[BaseException],
    exc_obj: BaseException,
    exc_tb: TracebackType,
) -> None:
    traceback.print_exc()


@deprecated("Use aiotools.TaskScope instead.")
class PersistentTaskGroup:
    _base_error: BaseException | None
    _exc_handler: AsyncExceptionHandler
    _tasks: set[asyncio.Task[Any]]
    _on_completed_fut: asyncio.Future[Any] | None
    _current_taskgroup_token: Token[PersistentTaskGroup] | None
    _entered: bool
    _exiting: bool
    _aborting: bool
    _name: str
    _parent_cancel_requested: bool
    _unfinished_tasks: int
    _parent_task: asyncio.Task[Any] | None

    @classmethod
    def all_ptaskgroups(cls) -> Sequence[PersistentTaskGroup]:
        return list(_all_ptaskgroups)

    def __init__(
        self,
        *,
        name: str | None = None,
        exception_handler: AsyncExceptionHandler | None = None,
    ) -> None:
        self._entered = False
        self._exiting = False
        self._aborting = False
        self._base_error = None
        self._name = name or f"{next(_ptaskgroup_idx)}"
        self._parent_cancel_requested = False
        self._unfinished_tasks = 0
        self._on_completed_fut = None
        self._parent_task = compat.current_task()
        self._tasks = set()
        self._current_taskgroup_token = None
        _all_ptaskgroups.add(self)
        if exception_handler is None:
            self._exc_handler = _default_exc_handler
        else:
            self._exc_handler = exception_handler

    # TODO: task statistics and enumeration (for aiomonitor)
    # TODO: two-phase shutdown

    def get_name(self) -> str:
        return self._name

    def create_task(
        self,
        coro: CoroutineLike[T],
        *,
        name: str | None = None,
    ) -> Awaitable[T]:
        if not self._entered:
            # When used as object attribute, auto-enter.
            self._entered = True
        if self._exiting and self._unfinished_tasks == 0:
            raise RuntimeError(f"{self!r} has already finished")
        return self._create_task_with_name(coro, name=name, cb=self._on_task_done)

    def _create_task_with_name(
        self,
        coro: CoroutineLike[T],
        *,
        name: str | None = None,
        cb: Callable[[asyncio.Task[T | None]], None],
    ) -> Awaitable[T]:
        loop = compat.get_running_loop()
        result_future = loop.create_future()
        child_task = loop.create_task(
            self._task_wrapper(coro, weakref.ref(result_future)),
            name=name,
        )
        _log.debug("%r is spawned in %r.", child_task, self)
        self._unfinished_tasks += 1
        child_task.add_done_callback(cb)
        self._tasks.add(child_task)
        child_task.add_done_callback(self._tasks.discard)
        return result_future

    def _is_base_error(self, exc: BaseException) -> bool:
        assert isinstance(exc, BaseException)
        return isinstance(exc, (SystemExit, KeyboardInterrupt))

    async def _wait_completion(self) -> BaseException | None:
        loop = compat.get_running_loop()
        propagate_cancellation_error: BaseException | None = None
        while self._unfinished_tasks:
            if self._on_completed_fut is None:
                self._on_completed_fut = loop.create_future()
            try:
                await self._on_completed_fut
            except asyncio.CancelledError as ex:
                if not self._aborting:
                    propagate_cancellation_error = ex
                    self._trigger_shutdown()
            self._on_completed_fut = None

        assert self._unfinished_tasks == 0
        self._on_completed_fut = None
        _all_ptaskgroups.discard(self)
        return propagate_cancellation_error

    def _trigger_shutdown(self) -> None:
        self._aborting = True
        for t in self._tasks:
            if not t.done():
                t.cancel()

    async def shutdown(self) -> None:
        self._trigger_shutdown()
        await self._wait_completion()

    async def _task_wrapper(
        self,
        coro: CoroutineLike[T],
        result_future: weakref.ref[asyncio.Future[T]],
    ) -> T | None:
        loop = compat.get_running_loop()
        task = compat.current_task()
        fut = result_future()
        try:
            ret = await cast(Awaitable[T], coro)  # forced cast for Python 3.12 or older
            if fut is not None:
                fut.set_result(ret)
            return ret
        except asyncio.CancelledError:
            if fut is not None:
                fut.cancel()
            raise
        except Exception as e:
            # Swallow unhandled exceptions by our own and
            # prevent abortion of the task group bu them.
            # Wrapping corotuines directly has advantage for
            # exception handlers to access full traceback
            # and there is no need to implement separate
            # mechanism to wait for exception handler tasks.
            try:
                if fut is not None:
                    fut.set_exception(e)
                exc_info = cast(
                    tuple[type[BaseException], BaseException, TracebackType],
                    sys.exc_info(),
                )
                await self._exc_handler(*exc_info)
            except Exception as exc:
                # If there are exceptions inside the exception handler
                # we report it as soon as possible using the event loop's
                # exception handler, instead of postponing
                # to the timing when PersistentTaskGroup terminates.
                loop.call_exception_handler({
                    "message": (
                        "Got an unhandled exception "
                        f"in the exception handler of Task {task!r}"
                    ),
                    "exception": exc,
                    "task": task,
                })
            return None
        finally:
            del fut

    def _on_task_done(self, task: asyncio.Task[Any]) -> None:
        self._unfinished_tasks -= 1
        assert self._unfinished_tasks >= 0
        assert self._parent_task is not None

        if self._on_completed_fut is not None and not self._unfinished_tasks:
            if not self._on_completed_fut.done():
                self._on_completed_fut.set_result(True)

        if task.cancelled():
            _log.debug("%r in %r has been cancelled.", task, self)
            return

        exc = task.exception()
        if exc is None:
            return

        # Now the exception is BaseException.
        if self._base_error is None:
            self._base_error = exc

        self._trigger_shutdown()
        if not self._parent_task.cancelling():
            self._parent_cancel_requested = True

    async def __aenter__(self) -> Self:
        self._parent_task = compat.current_task()
        self._current_taskgroup_token = current_ptaskgroup.set(self)
        self._entered = True
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool | None:
        assert self._parent_task is not None
        self._exiting = True
        propagate_cancellation_error: type[BaseException] | BaseException | None = None

        if (
            exc_val is not None
            and self._is_base_error(exc_val)
            and self._base_error is None
        ):
            self._base_error = exc_val

        if exc_type is asyncio.CancelledError or exc_type is asyncio.TimeoutError:
            if self._parent_cancel_requested:
                self._parent_task.uncancel()
            else:
                propagate_cancellation_error = exc_type
        if exc_type is not None and not self._aborting:
            if exc_type is asyncio.CancelledError or exc_type is asyncio.TimeoutError:
                propagate_cancellation_error = exc_type
            self._trigger_shutdown()

        prop_ex = await self._wait_completion()
        if prop_ex is not None:
            propagate_cancellation_error = prop_ex
        if self._current_taskgroup_token:
            current_ptaskgroup.reset(self._current_taskgroup_token)
            self._current_taskgroup_token = None
        if propagate_cancellation_error is not None:
            raise propagate_cancellation_error

        return None

    def __repr__(self) -> str:
        info = [""]
        if self._name:
            info.append(f"name={self._name}")
        if self._tasks:
            info.append(f"tasks={len(self._tasks)}")
        if self._unfinished_tasks:
            info.append(f"unfinished={self._unfinished_tasks}")
        if self._aborting:
            info.append("cancelling")
        elif self._entered:
            info.append("entered")
        info_str = " ".join(info)
        return f"<PersistentTaskGroup({info_str})>"



from __future__ import annotations

import asyncio
import multiprocessing as mp
import os
import signal
import sys
import time
from unittest import mock

import pytest

from aiotools import fork as fork_mod
from aiotools.fork import (
    AbstractChildProcess,
    MPContext,
    PidfdChildProcess,
    _has_pidfd,
    afork,
)

if sys.platform == "win32":
    pytest.skip(
        "forks tests not supported on Windows",
        allow_module_level=True,
    )


target_mp_contexts = [
    pytest.param(mp.get_context(method), id=method)
    for method in mp.get_all_start_methods()
    if method != "forkserver"
]


def child_for_fork() -> int:
    print("hello world")
    time.sleep(0.1)
    return 99


async def _do_test_fork(mp_context: MPContext) -> None:
    proc = await afork(child_for_fork, mp_context=mp_context)
    assert proc.pid > 0
    if isinstance(proc, PidfdChildProcess):
        assert proc._pidfd > 0
    ret = await proc.wait()
    assert ret == 99


def child_for_fork_already_terminated() -> int:
    time.sleep(0.1)
    return 99


async def _do_test_fork_already_terminated(mp_context: MPContext) -> None:
    proc = await afork(child_for_fork_already_terminated, mp_context=mp_context)
    assert proc.pid > 0
    if isinstance(proc, PidfdChildProcess):
        assert proc._pidfd > 0
    await asyncio.sleep(0.5)
    ret = await proc.wait()
    assert ret == 99


def child_for_fork_signal() -> int:
    try:
        time.sleep(10)
    except KeyboardInterrupt:
        return 101
    return 100


async def _do_test_fork_signal(mp_context: MPContext) -> None:
    os.setpgrp()
    proc = await afork(child_for_fork_signal, mp_context=mp_context)
    assert proc.pid > 0
    if isinstance(proc, PidfdChildProcess):
        assert proc._pidfd > 0
    await asyncio.sleep(0.1)
    proc.send_signal(signal.SIGINT)
    ret = await proc.wait()
    # FIXME: Sometimes it returns 254
    assert ret == 101


def child_for_fork_segfault() -> int:
    time.sleep(0.1)
    import ctypes

    ctypes.string_at(0)  # segfault!
    return 0


async def _do_test_fork_segfault(mp_context: MPContext) -> None:
    os.setpgrp()
    proc = await afork(child_for_fork_segfault, mp_context=mp_context)
    assert proc.pid > 0
    if isinstance(proc, PidfdChildProcess):
        assert proc._pidfd > 0
    ret = await proc.wait()
    assert ret == -11  # SIGSEGV


def child_for_fork_many() -> int:
    try:
        time.sleep(10)
    except KeyboardInterrupt:
        return 101
    return 100


async def _do_test_fork_many(mp_context: MPContext) -> None:
    os.setpgrp()
    proc_list: list[AbstractChildProcess] = []
    for _ in range(32):
        proc = await afork(child_for_fork_many, mp_context=mp_context)
        proc_list.append(proc)
        assert proc.pid > 0
        if isinstance(proc, PidfdChildProcess):
            assert proc._pidfd > 0
    for i in range(16):
        proc_list[i].send_signal(signal.SIGINT)
    for i in range(16, 32):
        proc_list[i].send_signal(signal.SIGTERM)
    ret_list = await asyncio.gather(*[proc.wait() for proc in proc_list])
    for i in range(16):
        assert ret_list[i] == 101
    for i in range(16, 32):
        assert ret_list[i] == -15  # killed by SIGTERM


@pytest.mark.skipif(
    not _has_pidfd, reason="pidfd is supported in Python 3.9+ and Linux kernel 5.4+"
)
@pytest.mark.parametrize("mp_context", target_mp_contexts)
@pytest.mark.asyncio
async def test_fork(mp_context: MPContext) -> None:
    await _do_test_fork(mp_context)


@pytest.mark.skipif(
    not _has_pidfd, reason="pidfd is supported in Python 3.9+ and Linux kernel 5.4+"
)
@pytest.mark.parametrize("mp_context", target_mp_contexts)
@pytest.mark.asyncio
async def test_fork_already_terminated(mp_context: MPContext) -> None:
    await _do_test_fork_already_terminated(mp_context)


@pytest.mark.skipif(
    not _has_pidfd, reason="pidfd is supported in Python 3.9+ and Linux kernel 5.4+"
)
@pytest.mark.parametrize("mp_context", target_mp_contexts)
@pytest.mark.asyncio
async def test_fork_signal(mp_context: MPContext) -> None:
    await _do_test_fork_signal(mp_context)


@pytest.mark.skipif(
    not _has_pidfd, reason="pidfd is supported in Python 3.9+ and Linux kernel 5.4+"
)
@pytest.mark.parametrize("mp_context", target_mp_contexts)
@pytest.mark.asyncio
async def test_fork_segfault(mp_context: MPContext) -> None:
    await _do_test_fork_segfault(mp_context)


@pytest.mark.skipif(
    not _has_pidfd, reason="pidfd is supported in Python 3.9+ and Linux kernel 5.4+"
)
@pytest.mark.parametrize("mp_context", target_mp_contexts)
@pytest.mark.asyncio
async def test_fork_many(mp_context: MPContext) -> None:
    await _do_test_fork_many(mp_context)


@pytest.mark.parametrize("mp_context", target_mp_contexts)
@pytest.mark.asyncio
async def test_fork_fallback(mp_context: MPContext) -> None:
    with mock.patch.object(
        fork_mod,
        "_has_pidfd",
        False,
    ):
        await _do_test_fork(mp_context)


@pytest.mark.parametrize("mp_context", target_mp_contexts)
@pytest.mark.asyncio
async def test_fork_already_termination_fallback(mp_context: MPContext) -> None:
    with mock.patch.object(
        fork_mod,
        "_has_pidfd",
        False,
    ):
        await _do_test_fork_already_terminated(mp_context)


@pytest.mark.parametrize("mp_context", target_mp_contexts)
@pytest.mark.asyncio
async def test_fork_signal_fallback(mp_context: MPContext) -> None:
    with mock.patch.object(
        fork_mod,
        "_has_pidfd",
        False,
    ):
        await _do_test_fork_signal(mp_context)


@pytest.mark.parametrize("mp_context", target_mp_contexts)
@pytest.mark.asyncio
async def test_fork_segfault_fallback(mp_context: MPContext) -> None:
    with mock.patch.object(
        fork_mod,
        "_has_pidfd",
        False,
    ):
        await _do_test_fork_segfault(mp_context)


@pytest.mark.parametrize("mp_context", target_mp_contexts)
@pytest.mark.asyncio
async def test_fork_many_fallback(mp_context: MPContext) -> None:
    with mock.patch.object(
        fork_mod,
        "_has_pidfd",
        False,
    ):
        await _do_test_fork_many(mp_context)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
import os
from pathlib import Path

import aiotools

on_rtd = os.environ.get("READTHEDOCS", None) == "True"
root_dir = Path(__file__).resolve().parents[1]

#
# aiotools documentation build configuration file, created by
# sphinx-quickstart on Tue Aug  1 15:09:44 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    # 'sphinx_autodoc_typehints',
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx_rtd_theme",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "aiotools"
copyright = "2018-2023, Joongi Kim and contributors"
author = "Joongi Kim and contributors"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

# The short X.Y version.
version = ".".join(aiotools.__version__.split(".")[:2])
# The full version, including alpha/beta/rc tags.
release = aiotools.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# import guzzle_sphinx_theme
# html_theme_path = guzzle_sphinx_theme.html_theme_path()
# extensions.append('guzzle_sphinx_theme')
html_theme = "sphinx_rtd_theme"
if not on_rtd:
    import sphinx_rtd_theme

    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]


# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "project_nav_name": "aiotools",
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
# html_sidebars = {}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "aiotoolsdoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements: dict[str, str] = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "aiotools.tex", "aiotools Documentation", "Joongi Kim", "manual"),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "aiotools", "aiotools Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "aiotools",
        "aiotools Documentation",
        author,
        "aiotools",
        "One line description of project.",
        "Miscellaneous",
    ),
]


add_module_names = False

intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}



import asyncio
from contextvars import ContextVar, copy_context
from typing import TypeVar

import pytest

from aiotools import VirtualClock, gather_safe

T = TypeVar("T")
cancelled = ContextVar("cancelled", default=0)


async def do_job(delay: float, result: T) -> T:
    try:
        await asyncio.sleep(delay)
        return result
    except asyncio.CancelledError:
        await asyncio.sleep(0.1)
        cancelled.set(cancelled.get() + 1)
        raise


async def fail_job(delay: float) -> None:
    try:
        await asyncio.sleep(delay)
        1 / 0
    except asyncio.CancelledError:
        await asyncio.sleep(0.1)
        cancelled.set(cancelled.get() + 1)
        raise


@pytest.mark.asyncio
async def test_gather_safe() -> None:
    context = copy_context()
    with VirtualClock().patch_loop():

        async def _inner() -> None:
            results = await gather_safe(
                [
                    do_job(0.3, 3),
                    do_job(0.2, 2),
                    do_job(0.1, 1),
                ],
                context=context,
            )
            assert results == [3, 2, 1]

        await asyncio.create_task(_inner(), context=context)


@pytest.mark.asyncio
async def test_gather_safe_partial_failure() -> None:
    context = copy_context()
    with VirtualClock().patch_loop():

        async def _inner() -> None:
            results = await gather_safe(
                [
                    do_job(0.3, 3),
                    fail_job(0.2),
                    do_job(0.1, 1),
                ],
                context=context,
            )
            assert results[0] == 3
            assert isinstance(results[1], ZeroDivisionError)
            assert results[2] == 1
            assert cancelled.get() == 0

        await asyncio.create_task(_inner(), context=context)


@pytest.mark.asyncio
async def test_gather_safe_timeout() -> None:
    context = copy_context()
    with VirtualClock().patch_loop():

        async def _inner() -> None:
            with pytest.raises(asyncio.TimeoutError):
                async with asyncio.timeout(0.35):
                    await gather_safe(
                        [
                            do_job(0.1, 1),
                            fail_job(0.2),
                            fail_job(0.25),
                            do_job(0.3, 3),
                            # timeout occurs here
                            do_job(0.4, 4),  # cancelled
                            fail_job(0.5),  # cancelled
                            fail_job(0.6),  # cancelled
                        ],
                        context=context,
                    )
            # There is no way to retrieve partial result in this case.
            # If you want to combine gather_safe() and timeouts, your
            # tasks should take care of any potential exceptions and results
            # by on their own, instead of relying on the return value of
            # the gather_safe() call.
            # We track how many were cancelled indirectly using a contextvar.
            assert cancelled.get() == 3

        await asyncio.create_task(_inner(), context=context)



# mypy: disable-error-code="deprecated"
# pyright: reportDeprecated=false

from importlib.metadata import version

__version__ = version("aiotools")

from .cancel import (
    cancel_and_wait,
)
from .compat import (
    all_tasks,
    current_task,
    get_running_loop,
    set_task_name,
)
from .context import (
    AsyncContextGroup,
    AsyncContextManager,
    AsyncExitStack,
    aclosing,
    actxgroup,
    actxmgr,
    async_ctx_manager,
    closing_async,
    resetting,
)
from .defer import (
    AsyncDeferFunc,
    AsyncDeferrable,
    DeferFunc,
    adefer,
    defer,
)
from .fork import (
    AbstractChildProcess,
    PidfdChildProcess,
    PosixChildProcess,
    afork,
)
from .func import (
    apartial,
    lru_cache,
)
from .iter import aiter
from .server import (
    AsyncServerContextManager,
    InterruptedBySignal,
    ServerMainContextManager,
    main_context,
    process_index,
    server_context,
    start_server,
)
from .supervisor import Supervisor
from .taskcontext import (
    ErrorArg,
    ErrorCallback,
    TaskContext,
)
from .taskgroup import (
    MultiError,
    PersistentTaskGroup,
    TaskGroup,
    TaskGroupError,
    current_ptaskgroup,
    current_taskgroup,
)
from .taskscope import TaskScope
from .timeouts import (
    Timeout,
    timeout,
    timeout_at,
)
from .timer import (
    TimerDelayPolicy,
    VirtualClock,
    create_timer,
)
from .types import (
    AsyncClosable,
    AwaitableLike,
    CoroutineLike,
)
from .utils import (
    as_completed_safe,
    gather_safe,
    race,
)

main = main_context

__all__ = (
    # .cancel
    "cancel_and_wait",
    # .compat
    "all_tasks",
    "get_running_loop",
    "current_task",
    "set_task_name",
    # .context
    "resetting",
    "AsyncContextManager",
    "async_ctx_manager",
    "actxmgr",
    "aclosing",
    "closing_async",
    "AsyncContextGroup",
    "actxgroup",
    "AsyncExitStack",
    # .defer
    "DeferFunc",
    "AsyncDeferrable",
    "AsyncDeferFunc",
    "adefer",
    "defer",
    # .fork
    "AbstractChildProcess",
    "PosixChildProcess",
    "PidfdChildProcess",
    "afork",
    # .func
    "apartial",
    "lru_cache",
    # .iter
    "aiter",
    # .server
    "main",
    "main_context",
    # NOTE: "@aiotools.server" still works,
    #       but server_context is provided to silence typecheckers.
    "server_context",
    "start_server",
    "process_index",
    "AsyncServerContextManager",
    "ServerMainContextManager",
    "InterruptedBySignal",
    # .supervisor
    "Supervisor",
    # .taskcontext
    "ErrorArg",
    "ErrorCallback",
    "TaskContext",
    # .taskgroup
    "MultiError",
    "TaskGroup",
    "TaskGroupError",
    "current_taskgroup",
    "PersistentTaskGroup",
    "current_ptaskgroup",
    # .taskscope
    "TaskScope",
    # .timeouts
    "Timeout",
    "timeout",
    "timeout_at",
    # .timer
    "TimerDelayPolicy",
    "VirtualClock",
    "create_timer",
    # .types
    "AsyncClosable",
    "AwaitableLike",
    "CoroutineLike",
    # .utils
    "as_completed_safe",
    "gather_safe",
    "race",
    "__version__",
)



from __future__ import annotations

import asyncio
import functools
import glob
import logging.config
import multiprocessing as mp
import os
import signal
import sys
import tempfile
import threading
import time
from collections.abc import AsyncGenerator, Callable, Generator, Iterator
from types import FrameType
from typing import Any, Optional, Sequence

import pytest

import aiotools
from aiotools.fork import MPContext

if os.environ.get("CI", "") and sys.version_info < (3, 9, 0):
    pytest.skip(
        "skipped to prevent kill CI agents due to signals on CI environments",
        allow_module_level=True,
    )

if sys.platform == "win32":
    pytest.skip(
        "server tests not supported on Windows",
        allow_module_level=True,
    )

target_mp_contexts = [
    pytest.param(mp.get_context(method), id=method)
    for method in mp.get_all_start_methods()
]
target_mp_contexts_without_forkserver = [
    pytest.param(mp.get_context(method), id=method)
    for method in mp.get_all_start_methods()
    if method != "forkserver"
]


@pytest.fixture
def restore_signal() -> Iterator[None]:
    os.setpgrp()
    old_alrm = signal.getsignal(signal.SIGALRM)
    old_intr = signal.getsignal(signal.SIGINT)
    old_term = signal.getsignal(signal.SIGTERM)
    old_intr = signal.getsignal(signal.SIGUSR1)
    yield
    signal.signal(signal.SIGALRM, old_alrm)
    signal.signal(signal.SIGINT, old_intr)
    signal.signal(signal.SIGTERM, old_term)
    signal.signal(signal.SIGUSR1, old_term)


@pytest.fixture
def set_timeout() -> Iterator[Callable[[float, Callable[..., None]], None]]:
    def make_timeout(sec: float, callback: Any) -> None:
        def _callback(signum: int, frame: FrameType | None) -> None:
            signal.alarm(0)
            callback()

        signal.signal(signal.SIGALRM, _callback)
        signal.setitimer(signal.ITIMER_REAL, sec)

    yield make_timeout


def write_record(record_name: str, msg: str) -> None:
    path = f"{record_name}.{os.getpid()}"
    with open(path, "a", encoding="utf8") as writer:
        writer.write(msg + "\n")


def read_records(record_name: str) -> Sequence[str]:
    lines: list[str] = []
    for path in glob.glob(f"{record_name}.*"):
        with open(path, "r", encoding="utf8") as reader:
            lines.extend(line.strip() for line in reader.readlines())
    return lines


@pytest.fixture
def exec_recorder() -> Iterator[str]:
    f = tempfile.NamedTemporaryFile(
        mode="w",
        encoding="utf8",
        prefix="aiotools.tests.server.",
    )
    f.close()

    yield f.name

    for path in glob.glob(f"{f.name}.*"):
        os.unlink(path)


def interrupt(pid: int = 0, signum: signal.Signals = signal.SIGINT) -> None:
    os.kill(pid, signum)


@aiotools.server_context
async def myserver_simple(
    loop: asyncio.AbstractEventLoop,
    proc_idx: int,
    args: Sequence[Any],
) -> AsyncGenerator[None, signal.Signals]:
    record_name = args[0]
    await asyncio.sleep(0)
    write_record(record_name, f"started:{proc_idx}")

    yield

    await asyncio.sleep(0)
    write_record(record_name, f"terminated:{proc_idx}")


@pytest.mark.parametrize("mp_context", target_mp_contexts)
def test_server_singleproc(
    restore_signal: None,
    exec_recorder: str,
    mp_context: MPContext,
) -> None:
    record_name = exec_recorder
    aiotools.start_server(
        myserver_simple,
        args=(record_name,),
        mp_context=mp_context,
        run_to_completion=True,
    )
    lines = set(read_records(record_name))
    assert "started:0" in lines
    assert "terminated:0" in lines


@pytest.mark.parametrize("mp_context", target_mp_contexts)
def test_server_multiproc(
    restore_signal: None,
    exec_recorder: str,
    mp_context: MPContext,
) -> None:
    record_name = exec_recorder
    aiotools.start_server(
        myserver_simple,
        num_workers=3,
        args=(record_name,),
        mp_context=mp_context,
        run_to_completion=True,
    )
    lines = set(read_records(record_name))
    assert lines == {
        "started:0",
        "started:1",
        "started:2",
        "terminated:0",
        "terminated:1",
        "terminated:2",
    }


@aiotools.server_context
async def myserver_signal(
    loop: asyncio.AbstractEventLoop,
    proc_idx: int,
    args: Sequence[Any],
) -> AsyncGenerator[None, signal.Signals]:
    record_name = args[0]
    await asyncio.sleep(0)
    write_record(record_name, f"started:{proc_idx}")

    received_signum = yield

    await asyncio.sleep(0)
    write_record(record_name, f"terminated:{proc_idx}:{received_signum}")


@pytest.mark.parametrize("mp_context", target_mp_contexts)
def test_server_multiproc_custom_stop_signals(
    restore_signal: None,
    exec_recorder: str,
    mp_context: MPContext,
) -> None:
    record_name = exec_recorder
    aiotools.start_server(
        myserver_signal,
        num_workers=2,
        stop_signals={signal.SIGUSR1, signal.SIGUSR2},
        args=(record_name,),
        mp_context=mp_context,
        # When custom stop signals are set, the run-to-completion mode
        # sets the yield's return value as the "first" stop signal by
        # the numerical ascending order.
        run_to_completion=True,
    )
    lines = set(read_records(record_name))
    assert {"started:0", "started:1"} < lines
    assert {
        f"terminated:0:{int(signal.SIGUSR1)}",
        f"terminated:1:{int(signal.SIGUSR1)}",
    } < lines


@aiotools.server_context
async def myserver_worker_init_error(
    loop: asyncio.AbstractEventLoop,
    proc_idx: int,
    args: Sequence[Any],
) -> AsyncGenerator[None, signal.Signals]:
    record_name = args[0]

    class _LogAdaptor:
        def __init__(self, writer: Any) -> None:
            self.writer = writer

        def write(self, msg: str) -> None:
            msg = msg.strip().replace("\n", " ")
            self.writer(f"log:{proc_idx}:{msg}")

    log_stream = _LogAdaptor(functools.partial(write_record, record_name))
    logging.config.dictConfig({
        "version": 1,
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "stream": log_stream,
                "level": "DEBUG",
            },
        },
        "loggers": {
            "aiotools": {
                "handlers": ["console"],
                "level": "DEBUG",
            },
        },
    })
    log = logging.getLogger("aiotools")
    write_record(record_name, f"started:{proc_idx}")
    log.debug("hello")
    if proc_idx in (0, 2):
        raise ZeroDivisionError("oops")

    yield

    # should not be reached if errored.
    await asyncio.sleep(0)
    write_record(record_name, f"terminated:{proc_idx}")


@pytest.mark.parametrize("mp_context", target_mp_contexts)
def test_server_worker_init_error(
    restore_signal: None,
    exec_recorder: str,
    mp_context: MPContext,
) -> None:
    record_name = exec_recorder
    aiotools.start_server(
        myserver_worker_init_error,
        num_workers=4,
        args=(record_name,),
        mp_context=mp_context,
        # Let it wait until all workers finish even when some of them raises unhandled exceptions
        # Otherwise, any first received worker-to-main interrupt message via the intr-pipe will cancel
        # the forever future and could cause race conditions.
        run_to_completion=True,
    )
    lines = set(read_records(record_name))
    assert sum(1 if line.startswith("started:") else 0 for line in lines) == 4
    # workers who did not raise errors have already started,
    # and they should have terminated normally
    # when the errorneous worker interrupted the main loop.
    assert sum(1 if line.startswith("terminated:") else 0 for line in lines) == 2
    assert sum(1 if "hello" in line else 0 for line in lines) == 4
    assert sum(1 if "ZeroDivisionError: oops" in line else 0 for line in lines) == 2


main_enter = False
main_exit = False
main_signal = 0


@aiotools.main_context
def mymain_user_main() -> Generator[int, signal.Signals]:
    # The main context manager is executed inside the main process.
    global main_enter, main_exit
    main_enter = True
    yield 987
    main_exit = True


@aiotools.server_context
async def myworker_user_main(
    loop: asyncio.AbstractEventLoop,
    proc_idx: int,
    args: Sequence[Any],
) -> AsyncGenerator[None, signal.Signals]:
    assert args[0] == 987  # first arg from user main
    assert args[1] == 123  # second arg from start_server args
    yield


@pytest.mark.parametrize("mp_context", target_mp_contexts)
def test_server_user_main(
    restore_signal: None,
    mp_context: MPContext,
) -> None:
    global main_enter, main_exit
    main_enter = False
    main_exit = False
    aiotools.start_server(
        myworker_user_main,
        mymain_user_main,
        num_workers=3,
        args=(123,),
        mp_context=mp_context,
        run_to_completion=True,
    )
    assert main_enter
    assert main_exit


@aiotools.main_context
def mymain_for_custom_stop_signals() -> Generator[None, signal.Signals]:
    # The main context manager is executed inside the main process.
    global main_enter, main_exit, main_signal
    main_enter = True
    main_signal = yield
    main_exit = True


@aiotools.server_context
async def myworker_for_custom_stop_signals(
    loop: asyncio.AbstractEventLoop,
    proc_idx: int,
    args: Sequence[Any],
) -> AsyncGenerator[None, signal.Signals]:
    worker_signals = args[0]
    worker_signals[proc_idx] = yield


@pytest.mark.parametrize("mp_context", target_mp_contexts)
def test_server_user_main_custom_stop_signals(
    restore_signal: None,
    mp_context: MPContext,
) -> None:
    global main_enter, main_exit, main_signal
    main_enter = False
    main_exit = False
    main_signal = 0
    worker_signals = mp_context.Array("i", 3)

    aiotools.start_server(
        myworker_for_custom_stop_signals,
        mymain_for_custom_stop_signals,
        num_workers=3,
        stop_signals={signal.SIGUSR1},
        args=(worker_signals,),
        mp_context=mp_context,
        run_to_completion=True,
    )

    assert main_enter
    assert main_exit
    assert main_signal == signal.SIGUSR1
    assert worker_signals[0] == signal.SIGUSR1
    assert worker_signals[1] == signal.SIGUSR1
    assert worker_signals[2] == signal.SIGUSR1


@aiotools.main_context
def mymain_for_main_tuple() -> Generator[tuple[int, int], signal.Signals]:
    global main_enter, main_exit
    main_enter = True
    yield 987, 654
    main_exit = True


@aiotools.server_context
async def myworker_for_main_tuple(
    loop: asyncio.AbstractEventLoop,
    proc_idx: int,
    args: Sequence[Any],
) -> AsyncGenerator[None, signal.Signals]:
    assert args[0] == 987  # first arg from user main
    assert args[1] == 654  # second arg from user main
    assert args[2] == 123  # third arg from start_server args
    yield


@pytest.mark.parametrize("mp_context", target_mp_contexts)
def test_server_user_main_tuple(
    restore_signal: None,
    mp_context: MPContext,
) -> None:
    global main_enter, main_exit
    main_enter = False
    main_exit = False

    aiotools.start_server(
        myworker_for_main_tuple,
        mymain_for_main_tuple,
        num_workers=3,
        args=(123,),
        mp_context=mp_context,
        run_to_completion=True,
    )

    assert main_enter
    assert main_exit


@aiotools.server_context
async def myworker_for_extra_proc(
    loop: asyncio.AbstractEventLoop,
    proc_idx: int,
    args: Sequence[Any],
) -> AsyncGenerator[None, signal.Signals]:
    print(f"extra_proc_worker({proc_idx=})")
    yield


def extra_proc_plain(
    key: int,
    /,
    intr_event: Optional[threading.Event],
    pidx: int,
    args: Sequence[Any],
) -> None:
    assert intr_event is None  # unused with multiprocessing mode
    extras = args[0]
    extras[key] = 980 + key
    try:
        time.sleep(0.1)
    except KeyboardInterrupt:
        print(f"extra[{key}] interrupted", file=sys.stderr)
    except Exception as e:
        print(f"extra[{key}] exception", e, file=sys.stderr)
    finally:
        print(f"extra[{key}] finish", file=sys.stderr)
        extras[key] = 990 + key


@pytest.mark.parametrize("mp_context", target_mp_contexts)
def test_server_extra_proc(restore_signal: None, mp_context: MPContext) -> None:
    extras = mp_context.Array("i", [0, 0, 0])
    aiotools.start_server(
        myworker_for_extra_proc,
        extra_procs=[
            functools.partial(extra_proc_plain, 0),
            functools.partial(extra_proc_plain, 1),
            functools.partial(extra_proc_plain, 2),
        ],
        num_workers=3,
        args=(extras,),
        mp_context=mp_context,
        run_to_completion=True,
    )

    assert extras[0] == 990
    assert extras[1] == 991
    assert extras[2] == 992


@aiotools.server_context
async def myworker_with_extra_proc_for_custom_stop_signal(
    loop: asyncio.AbstractEventLoop,
    proc_idx: int,
    args: Sequence[Any],
) -> AsyncGenerator[None, signal.Signals]:
    yield


def extra_proc_for_custom_stop_signal(
    key: int,
    /,
    intr_event: Optional[threading.Event],
    pidx: int,
    args: Sequence[Any],
) -> None:
    assert intr_event is None  # unused with multiprocessing mode
    received_signals = args[0]
    try:
        time.sleep(10)
    except aiotools.InterruptedBySignal as e:
        received_signals[key] = e.args[0]


@pytest.mark.parametrize("mp_context", target_mp_contexts_without_forkserver)
def test_server_extra_proc_custom_stop_signal(
    set_timeout: Callable[[float, Callable[..., None]], None],
    restore_signal: None,
    mp_context: MPContext,
) -> None:
    # In local tests, the timeout may be as short as 0.x seconds,
    # but in GitHub Actions, we should assume more than 1 seconds of delay
    # for each worker process spawned.

    # Known issue:
    # The forkserver frequently hangs with custom stop signals
    # as multirpocessing resource tracker is not protected from them.
    set_timeout(3.0, functools.partial(interrupt, signum=signal.SIGUSR1))
    received_signals = mp_context.Array("i", [0, 0])
    aiotools.start_server(
        myworker_with_extra_proc_for_custom_stop_signal,
        extra_procs=[
            functools.partial(extra_proc_for_custom_stop_signal, 0),
            functools.partial(extra_proc_for_custom_stop_signal, 1),
        ],
        stop_signals={signal.SIGUSR1},
        args=(received_signals,),
        num_workers=3,
        mp_context=mp_context,
    )

    # This test case will generate a warning:
    #
    #   UserWarning: resource_tracker: process died unexpectedly, relaunching. Some resources might leak.
    #
    # because multiprocessing's resource tracker process gets killed
    # by the signal other than SIGINT or SIGTERM.
    # The list of blocked signals are hard-coded in the stdlib,
    # so there is nothing we can do here. :(
    assert received_signals[0] == signal.SIGUSR1
    assert received_signals[1] == signal.SIGUSR1



from __future__ import annotations

import contextvars

from .taskscope import TaskScope

__all__ = ("Supervisor",)


class Supervisor(TaskScope):
    """
    Supervisor is a primitive structure to provide a long-lived context manager scope
    for an indefinite set of subtasks.  During its lifetime, it is free to spawn new
    subtasks at any time.  If the supervisor itself is cancelled from outside or
    :meth:`aclose()` is called, it will cancel all running tasks immediately, wait
    for their completion, and then exit the context manager block.

    The main difference to :class:`asyncio.TaskGroup` is that it keeps running
    sibling subtasks even when there is an unhandled exception from one of the
    subtasks.

    To prevent memory leaks, a supervisor does not store any result or exception
    from its subtasks.  Instead, the callers must use additional task-done
    callbacks to process subtask results and exceptions.

    Supervisor provides the same analogy to Kotlin's |SupervisorScope|_ and
    Javascript's |Promise.allSettled()|_, while :class:`asyncio.TaskGroup` provides
    the same analogy to Kotlin's |CoroutineScope|_ and Javascript's |Promise.all()|_.

    .. |SupervisorScope| replace:: ``SupervisorScope``
    .. |CoroutineScope| replace:: ``CoroutineScope``
    .. _SupervisorScope: https://kotlinlang.org/api/kotlinx.coroutines/kotlinx-coroutines-core/kotlinx.coroutines/supervisor-scope.html
    .. _CoroutineScope: https://kotlinlang.org/api/kotlinx.coroutines/kotlinx-coroutines-core/kotlinx.coroutines/-coroutine-scope/
    .. |Promise.allSettled()| replace:: ``Promise.allSettled()``
    .. |Promise.all()| replace:: ``Promise.all()``
    .. _Promise.allSettled(): https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise/allSettled
    .. _Promise.all(): https://developer.mozilla.org/ko/docs/Web/JavaScript/Reference/Global_Objects/Promise/all

    The original implementation is based on DontPanicO's pull request
    (https://github.com/achimnol/cpython/pull/31) and :class:`PersistentTaskGroup`,
    but it is modified *not* to store unhandled subtask exceptions.

    .. versionadded:: 2.0
    """  # noqa: E501

    def __init__(self, context: contextvars.Context | None = None) -> None:
        super().__init__(exception_handler=None, context=context)



from __future__ import annotations

import sys
from collections.abc import Awaitable, Coroutine, Generator
from types import TracebackType
from typing import (
    Any,
    Protocol,
    TypeAlias,
    TypeVar,
    runtime_checkable,
)

__all__ = (
    "AsyncClosable",
    "AwaitableLike",
    "CoroutineLike",
    "ExcInfo",
    "OptExcInfo",
)

_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)


@runtime_checkable
class AsyncClosable(Protocol):
    async def close(self) -> None: ...


# Exception info type aliases for __aexit__ and similar methods
ExcInfo: TypeAlias = tuple[type[BaseException], BaseException, TracebackType]
OptExcInfo: TypeAlias = ExcInfo | tuple[None, None, None]

# taken from the typeshed
if sys.version_info >= (3, 12):
    AwaitableLike: TypeAlias = Awaitable[_T_co]  # noqa: Y047
    CoroutineLike: TypeAlias = Coroutine[Any, Any, _T_co]  # noqa: Y047
else:
    AwaitableLike: TypeAlias = Generator[Any, None, _T_co] | Awaitable[_T_co]
    CoroutineLike: TypeAlias = Generator[Any, None, _T_co] | Coroutine[Any, Any, _T_co]



from __future__ import annotations

import asyncio
from typing import Any, TypeVar

import pytest

from aiotools import Supervisor, TaskScope, VirtualClock

T = TypeVar("T")


async def do_job(delay: float, result: T) -> T:
    await asyncio.sleep(delay)
    return result


async def fail_job(delay: float) -> None:
    await asyncio.sleep(delay)
    raise ZeroDivisionError()


@pytest.mark.asyncio
async def test_taskscope_keep_running() -> None:
    results: list[str] = []

    async def parent() -> None:
        async with TaskScope() as ts:
            ts.create_task(fail_job(0.2))
            await asyncio.sleep(0.5)
            results.append("context-final")  # kept running
        await asyncio.sleep(0.5)
        results.append("parent-final")  # unaffected

    with VirtualClock().patch_loop():
        parent_task = asyncio.create_task(parent())
        await parent_task
        assert results == ["context-final", "parent-final"]


@pytest.mark.asyncio
async def test_supervisor_partial_failure() -> None:
    results: list[int] = []
    errors: list[BaseException] = []
    tasks: list[asyncio.Task[Any]] = []
    with VirtualClock().patch_loop():
        async with Supervisor() as supervisor:
            tasks.append(supervisor.create_task(do_job(0.1, 1)))
            tasks.append(supervisor.create_task(fail_job(0.2)))
            tasks.append(supervisor.create_task(do_job(0.3, 3)))
        for t in tasks:
            try:
                results.append(t.result())
            except Exception as e:
                errors.append(e)
    assert results == [1, 3]
    assert len(errors) == 1
    assert isinstance(errors[0], ZeroDivisionError)


@pytest.mark.asyncio
async def test_supervisor_timeout_before_failure() -> None:
    results: list[int] = []
    errors: list[BaseException] = []
    cancelled = 0
    tasks: list[asyncio.Task[Any]] = []
    with VirtualClock().patch_loop():
        with pytest.raises(TimeoutError):
            async with (
                asyncio.timeout(0.15),
                Supervisor() as supervisor,
            ):
                tasks.append(supervisor.create_task(do_job(0.1, 1)))
                # timeout here
                tasks.append(supervisor.create_task(fail_job(0.2)))
                tasks.append(supervisor.create_task(do_job(0.3, 3)))
        for t in tasks:
            try:
                results.append(t.result())
            except asyncio.CancelledError:
                cancelled += 1
            except Exception as e:
                errors.append(e)
    assert results == [1]
    assert len(errors) == 0
    assert cancelled == 2


@pytest.mark.asyncio
async def test_supervisor_timeout_after_failure() -> None:
    results: list[int] = []
    errors: list[BaseException] = []
    cancelled = 0
    tasks: list[asyncio.Task[Any]] = []
    with VirtualClock().patch_loop():
        with pytest.raises(TimeoutError):
            async with (
                asyncio.timeout(0.25),
                Supervisor() as supervisor,
            ):
                tasks.append(supervisor.create_task(do_job(0.1, 1)))
                tasks.append(supervisor.create_task(fail_job(0.2)))
                # timeout here
                tasks.append(supervisor.create_task(do_job(0.3, 3)))
        for t in tasks:
            try:
                results.append(t.result())
            except asyncio.CancelledError:
                cancelled += 1
            except Exception as e:
                errors.append(e)
    assert results == [1]
    assert len(errors) == 1
    assert isinstance(errors[0], ZeroDivisionError)
    assert cancelled == 1



from __future__ import annotations

import asyncio
from collections.abc import AsyncIterable, Iterator

import pytest

import aiotools


def test_iter() -> None:
    # extra test for cross-check with builtin iter()

    def stream() -> Iterator[int]:
        for i in range(10):
            yield i

    last_x = -1
    for x in iter(stream()):
        last_x = x

    assert last_x == 9


def test_iter_sentinel() -> None:
    # extra test for cross-check with builtin iter()

    _call_count = 0
    _sentinel = 5

    def get_next() -> int:
        nonlocal _call_count
        _call_count += 1
        return _call_count

    last_x = -1
    for x in iter(get_next, _sentinel):
        last_x = x

    assert last_x == _sentinel - 1


@pytest.mark.asyncio
async def test_aiter() -> None:
    async def stream() -> AsyncIterable[int]:
        for i in range(10):
            yield i

    last_x = -1
    async for x in aiotools.aiter(stream()):
        last_x = x

    assert last_x == 9


@pytest.mark.asyncio
async def test_aiter_with_sentinel() -> None:
    _call_count = 0
    _sentinel = 5

    async def get_next() -> int:
        nonlocal _call_count
        _call_count += 1
        await asyncio.sleep(0.001)
        return _call_count

    last_x = -1
    async for x in aiotools.aiter(get_next, _sentinel):
        last_x = x

    assert last_x == _sentinel - 1


@pytest.mark.asyncio
async def test_aiter_with_null_sentinel() -> None:
    _call_count = 0
    _sentinel = 3

    async def get_next() -> int | None:
        nonlocal _call_count
        _call_count += 1
        await asyncio.sleep(0.001)
        if _call_count >= _sentinel:
            return None
        return _call_count

    last_x: int | None = -1
    async for x in aiotools.aiter(get_next, None):
        last_x = x

    assert last_x == _sentinel - 1



import zmq

if __name__ == "__main__":
    ctx = zmq.Context()
    s = ctx.socket(zmq.PUSH)
    s.connect("tcp://127.0.0.1:5033")

    for _ in range(100):
        s.send(b"hello world")

    s.close()



from __future__ import annotations

import asyncio
import sys
import warnings
from collections.abc import AsyncGenerator, AsyncIterator
from contextlib import suppress
from contextvars import ContextVar

import pytest

import aiotools
from aiotools.context import (
    AbstractAsyncContextManager,
    resetting,
)

my_variable: ContextVar[int] = ContextVar("my_variable")


def test_resetting_ctxvar() -> None:
    with pytest.raises(LookupError):
        my_variable.get()
    with resetting(my_variable, 1):
        assert my_variable.get() == 1
        with resetting(my_variable, 2):
            assert my_variable.get() == 2
        assert my_variable.get() == 1
    with pytest.raises(LookupError):
        my_variable.get()

    # should behave the same way even when an exception occurs
    with suppress(RuntimeError):
        with resetting(my_variable, 10):
            assert my_variable.get() == 10
            raise RuntimeError("oops")
    with pytest.raises(LookupError):
        my_variable.get()


@pytest.mark.asyncio
async def test_resetting_ctxvar_async() -> None:
    with pytest.raises(LookupError):
        my_variable.get()
    async with resetting(my_variable, 1):
        assert my_variable.get() == 1
        async with resetting(my_variable, 2):
            assert my_variable.get() == 2
        assert my_variable.get() == 1
    with pytest.raises(LookupError):
        my_variable.get()

    # should behave the same way even when an exception occurs
    with suppress(RuntimeError):
        async with resetting(my_variable, 10):
            assert my_variable.get() == 10
            raise RuntimeError("oops")
    with pytest.raises(LookupError):
        my_variable.get()


def test_actxmgr_types() -> None:
    from typing import Any

    assert issubclass(aiotools.AsyncContextManager, AbstractAsyncContextManager)

    class boilerplate_ctx:
        async def __aenter__(self) -> Any:
            return self

        async def __aexit__(self, *exc_info: Any) -> None:
            return None

    @aiotools.actxmgr
    async def simple_ctx() -> AsyncIterator[None]:
        yield

    async def simple_agen() -> AsyncGenerator[None]:
        yield

    assert issubclass(boilerplate_ctx, AbstractAsyncContextManager)
    assert isinstance(simple_ctx(), AbstractAsyncContextManager)
    assert not isinstance(simple_agen(), AbstractAsyncContextManager)


@pytest.mark.asyncio
async def test_actxmgr() -> None:
    step = 0

    @aiotools.actxmgr
    async def simple_ctx(msg: str) -> AsyncIterator[str]:
        nonlocal step
        step = 1
        await asyncio.sleep(0)
        step = 2
        try:
            yield msg
            step = 3
        finally:
            await asyncio.sleep(0)
            step = 4

    step = 0
    async with simple_ctx("hello") as msg:
        assert step == 2
        assert msg == "hello"
    assert step == 4

    step = 0
    try:
        async with simple_ctx("world") as msg:
            assert step == 2
            assert msg == "world"
            await asyncio.sleep(0)
            raise ValueError("something wrong")
    except Exception as e:
        await asyncio.sleep(0)
        assert e.args[0] == "something wrong"
        assert step == 4


@pytest.mark.skipif(
    sys.version_info >= (3, 7, 0), reason="Deprecated in Python 3.7 or higher"
)
@pytest.mark.asyncio
async def test_actxmgr_reuse() -> None:
    @aiotools.actxmgr
    async def simple_ctx(msg: str) -> AsyncIterator[str]:
        yield msg

    cm = simple_ctx("hello")

    async with cm as msg:
        assert msg == "hello"

    try:
        async with cm as msg:
            assert msg == "hello"
    except BaseException as exc:
        assert isinstance(exc, RuntimeError)
        assert "didn't yield" in exc.args[0]

    cm = cm._recreate_cm()

    async with cm as msg:
        assert msg == "hello"


@pytest.mark.asyncio
async def test_actxmgr_exception_in_context_body() -> None:
    # Exceptions raised in the context body
    # should be transparently raised.

    @aiotools.actxmgr
    async def simple_ctx(msg: str) -> AsyncIterator[str]:
        await asyncio.sleep(0)
        yield msg
        await asyncio.sleep(0)

    with pytest.raises(ZeroDivisionError):
        async with simple_ctx("hello") as msg:
            assert msg == "hello"
            raise ZeroDivisionError

    exc = RuntimeError("oops")
    try:
        async with simple_ctx("hello") as msg:
            assert msg == "hello"
            raise exc
    except BaseException as e:
        assert e is exc
        assert e.args[0] == "oops"
    else:
        pytest.fail()

    cm = simple_ctx("hello")
    ret = await cm.__aenter__()
    assert ret == "hello"
    ret = await cm.__aexit__(ValueError, None, None)
    assert not ret


@pytest.mark.asyncio
async def test_actxmgr_exception_in_initialization() -> None:
    # Any exception before first yield is just transparently
    # raised out to the context block.

    @aiotools.actxmgr
    async def simple_ctx1(msg: str) -> AsyncIterator[str]:
        await asyncio.sleep(0)
        raise ZeroDivisionError
        await asyncio.sleep(0)
        yield msg
        await asyncio.sleep(0)

    try:
        async with simple_ctx1("hello") as msg:
            assert msg == "hello"
    except ZeroDivisionError:
        pass
    else:
        pytest.fail()

    exc = RuntimeError("oops")

    @aiotools.actxmgr
    async def simple_ctx2(msg: str) -> AsyncIterator[str]:
        await asyncio.sleep(0)
        raise exc
        await asyncio.sleep(0)
        yield msg
        await asyncio.sleep(0)

    try:
        async with simple_ctx2("hello") as msg:
            assert msg == "hello"
    except BaseException as e:
        assert e is exc
        assert e.args[0] == "oops"
    else:
        pytest.fail()


@pytest.mark.asyncio
async def test_actxmgr_exception_in_finalization() -> None:
    @aiotools.actxmgr
    async def simple_ctx1(msg: str) -> AsyncIterator[str]:
        await asyncio.sleep(0)
        yield msg
        await asyncio.sleep(0)
        raise ZeroDivisionError
        await asyncio.sleep(0)

    try:
        async with simple_ctx1("hello") as msg:
            assert msg == "hello"
    except ZeroDivisionError:
        pass
    else:
        pytest.fail()

    exc = RuntimeError("oops")

    @aiotools.actxmgr
    async def simple_ctx2(msg: str) -> AsyncIterator[str]:
        yield msg
        raise exc

    try:
        async with simple_ctx2("hello") as msg:
            assert msg == "hello"
    except BaseException as e:
        assert e is exc
        assert e.args[0] == "oops"
    else:
        pytest.fail()


@pytest.mark.asyncio
async def test_actxmgr_exception_uncaught() -> None:
    @aiotools.actxmgr
    async def simple_ctx(msg: str) -> AsyncIterator[str]:
        await asyncio.sleep(0)
        yield msg
        await asyncio.sleep(0)

    try:
        async with simple_ctx("hello"):
            raise IndexError("bomb")
    except BaseException as e:
        assert isinstance(e, IndexError)
        assert e.args[0] == "bomb"


@pytest.mark.asyncio
async def test_actxmgr_exception_nested() -> None:
    @aiotools.actxmgr
    async def simple_ctx(msg: str) -> AsyncIterator[str]:
        yield msg

    try:
        async with simple_ctx("hello") as msg1:
            async with simple_ctx("world") as msg2:
                assert msg1 == "hello"
                assert msg2 == "world"
                raise IndexError("bomb1")
    except BaseException as exc:
        assert isinstance(exc, IndexError)
        assert "bomb1" == exc.args[0]
    else:
        pytest.fail()


@pytest.mark.asyncio
async def test_actxmgr_exception_chained() -> None:
    @aiotools.actxmgr
    async def simple_ctx(msg: str) -> AsyncIterator[str]:
        try:
            await asyncio.sleep(0)
            yield msg
        except Exception as e:
            await asyncio.sleep(0)
            # exception is chained
            raise ValueError("bomb2") from e

    try:
        async with simple_ctx("hello") as msg:
            assert msg == "hello"
            raise IndexError("bomb1")
    except BaseException as exc:
        assert isinstance(exc, ValueError)
        assert "bomb2" == exc.args[0]
        assert isinstance(exc.__cause__, IndexError)
        assert "bomb1" == exc.__cause__.args[0]
    else:
        pytest.fail()


@pytest.mark.asyncio
async def test_actxmgr_exception_replaced() -> None:
    @aiotools.actxmgr
    async def simple_ctx(msg: str) -> AsyncIterator[str]:
        try:
            await asyncio.sleep(0)
            yield msg
        except Exception:
            await asyncio.sleep(0)
            # exception is replaced
            raise ValueError("bomb2")

    try:
        async with simple_ctx("hello") as msg:
            assert msg == "hello"
            raise IndexError("bomb1")
    except BaseException as exc:
        assert isinstance(exc, ValueError)
        assert "bomb2" == exc.args[0]
        assert exc.__cause__ is None
    else:
        pytest.fail()


@pytest.mark.asyncio
async def test_actxmgr_stopaiter() -> None:
    @aiotools.actxmgr
    async def simple_ctx1() -> AsyncIterator[int]:
        await asyncio.sleep(0)
        yield 1
        await asyncio.sleep(0)
        yield 2
        await asyncio.sleep(0)

    cm = simple_ctx1()
    ret_enter = await cm.__aenter__()
    assert ret_enter == 1
    ret_exit = await cm.__aexit__(StopAsyncIteration, None, None)
    assert ret_exit is False

    step = 0

    @aiotools.actxmgr
    async def simple_ctx2() -> AsyncIterator[None]:
        nonlocal step
        step = 1
        try:
            await asyncio.sleep(0)
            yield
        finally:
            step = 2
            await asyncio.sleep(0)
            raise StopAsyncIteration("x")

    try:
        step = 0
        async with simple_ctx2():
            assert step == 1
    except RuntimeError:
        # converted to RuntimeError
        assert step == 2
    else:
        pytest.fail()


@pytest.mark.asyncio
async def test_actxmgr_transparency() -> None:
    step = 0

    @aiotools.actxmgr
    async def simple_ctx1() -> AsyncIterator[None]:
        nonlocal step
        step = 1
        await asyncio.sleep(0)
        yield
        step = 2
        await asyncio.sleep(0)

    exc = StopAsyncIteration("x")
    try:
        step = 0
        async with simple_ctx1():
            assert step == 1
            raise exc
    except StopAsyncIteration as e:
        # exception is not handled
        assert e is exc
        assert step == 1
    else:
        pytest.fail()

    @aiotools.actxmgr
    async def simple_ctx2() -> AsyncIterator[None]:
        nonlocal step
        step = 1
        await asyncio.sleep(0)
        yield
        step = 2
        await asyncio.sleep(0)

    try:
        step = 0
        async with simple_ctx2():
            assert step == 1
            raise ValueError
    except ValueError:
        # exception is not handled
        assert step == 1
    else:
        pytest.fail()

    @aiotools.actxmgr
    async def simple_ctx3() -> AsyncIterator[None]:
        nonlocal step
        step = 1
        await asyncio.sleep(0)
        try:
            yield
        finally:
            step = 2
            await asyncio.sleep(0)

    exc = StopAsyncIteration("x")
    try:
        step = 0
        async with simple_ctx3():
            assert step == 1
            raise exc
    except StopAsyncIteration as e:
        # exception is not handled,
        # but context is finalized
        assert e is exc
        assert step == 2
    else:
        pytest.fail()

    @aiotools.actxmgr
    async def simple_ctx4() -> AsyncIterator[None]:
        nonlocal step
        step = 1
        await asyncio.sleep(0)
        try:
            yield
        finally:
            step = 2
            await asyncio.sleep(0)

    try:
        step = 0
        async with simple_ctx4():
            assert step == 1
            raise ValueError
    except ValueError:
        # exception is not handled,
        # but context is finalized
        assert step == 2
    else:
        pytest.fail()


@pytest.mark.asyncio
async def test_actxmgr_no_stop() -> None:
    @aiotools.actxmgr
    async def simple_ctx1(msg: str) -> AsyncIterator[str]:
        await asyncio.sleep(0)
        yield msg
        await asyncio.sleep(0)
        yield msg
        await asyncio.sleep(0)

    try:
        async with simple_ctx1("hello") as msg:
            assert msg == "hello"
    except RuntimeError as exc:
        assert "didn't stop" in exc.args[0]
    else:
        pytest.fail()

    try:
        async with simple_ctx1("hello") as msg:
            assert msg == "hello"
            raise ValueError("oops")
    except ValueError as exc:
        assert exc.args[0] == "oops"
    else:
        pytest.fail()

    @aiotools.actxmgr
    async def simple_ctx2(msg: str) -> AsyncIterator[str]:
        try:
            await asyncio.sleep(0)
            yield msg
        finally:
            await asyncio.sleep(0)
            yield msg

    try:
        async with simple_ctx2("hello") as msg:
            assert msg == "hello"
            raise ValueError("oops")
    except RuntimeError as exc:
        assert "didn't stop after" in exc.args[0]
    else:
        pytest.fail()


@pytest.mark.asyncio
async def test_actxmgr_no_yield() -> None:
    @aiotools.actxmgr  # type: ignore[arg-type]
    async def no_yield_ctx1(msg: str) -> AsyncIterator[None]:  # type: ignore[empty-body]
        pass

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        try:
            async with no_yield_ctx1("hello"):
                pass
        except RuntimeError as exc:
            assert "must be an async-gen" in exc.args[0]
        except AttributeError:  # in Python 3.7
            pass
        except TypeError as exc:  # in Python 3.10
            assert "not an async iterator" in exc.args[0]
        else:
            pytest.fail()


@pytest.mark.asyncio
async def test_actxgroup_exception_from_cm() -> None:
    @aiotools.actxmgr
    async def ctx1(a: int) -> AsyncIterator[int]:
        await asyncio.sleep(0)
        raise asyncio.CancelledError
        await asyncio.sleep(0)
        yield a

    @aiotools.actxmgr
    async def ctx2(a: int) -> AsyncIterator[int]:
        await asyncio.sleep(0)
        raise ZeroDivisionError
        await asyncio.sleep(0)
        yield a

    ctxgrp = aiotools.actxgroup([ctx1(1), ctx2(2)])

    async with ctxgrp as values:
        assert isinstance(values[0], asyncio.CancelledError)
        assert isinstance(values[1], ZeroDivisionError)

    @aiotools.actxmgr
    async def ctx3(a: int) -> AsyncIterator[int]:
        yield a
        raise asyncio.CancelledError

    @aiotools.actxmgr
    async def ctx4(a: int) -> AsyncIterator[int]:
        yield a
        raise ZeroDivisionError

    ctxgrp = aiotools.actxgroup([ctx3(1), ctx4(2)])

    async with ctxgrp as values:
        assert values[0] == 1
        assert values[1] == 2

    exits = ctxgrp.exit_states()
    assert isinstance(exits[0], asyncio.CancelledError)
    assert isinstance(exits[1], ZeroDivisionError)


@pytest.mark.asyncio
async def test_actxgroup_exception_from_body() -> None:
    exit_count = 0

    @aiotools.actxmgr
    async def ctx1(a: int) -> AsyncIterator[int]:
        nonlocal exit_count
        await asyncio.sleep(0)
        yield a
        await asyncio.sleep(0)
        # yield raises the exception from the context body.
        # If not handled, finalization will not be executed.
        exit_count += 1

    ctxgrp = aiotools.actxgroup([ctx1(1), ctx1(2)])

    try:
        async with ctxgrp as values:
            assert values[0] == 1
            assert values[1] == 2
            raise ZeroDivisionError
    except Exception as e:
        assert isinstance(e, ZeroDivisionError)

    exits = ctxgrp.exit_states()
    assert not exits[0]  # __aexit__ are called successfully
    assert not exits[1]
    assert exit_count == 0  # but they errored internally

    exit_count = 0

    @aiotools.actxmgr
    async def ctx2(a: int) -> AsyncIterator[int]:
        nonlocal exit_count
        try:
            await asyncio.sleep(0)
            yield a
        finally:
            await asyncio.sleep(0)
            # Ensure finalization is executed.
            exit_count += 1

    ctxgrp = aiotools.actxgroup([ctx2(1), ctx2(2)])

    try:
        async with ctxgrp as values:
            assert values[0] == 1
            assert values[1] == 2
            raise ZeroDivisionError
    except Exception as e:
        assert isinstance(e, ZeroDivisionError)

    exits = ctxgrp.exit_states()
    assert not exits[0]  # __aexit__ are called successfully
    assert not exits[1]
    assert exit_count == 2  # they also suceeeded


@pytest.mark.asyncio
async def test_aclosing() -> None:
    finalized = False

    async def myiter() -> AsyncGenerator[int]:
        nonlocal finalized
        try:
            yield 1
            await asyncio.sleep(0.2)
            yield 10
            await asyncio.sleep(0.2)
            yield 100
        except asyncio.CancelledError:
            await asyncio.sleep(0.2)
            finalized = True
            raise

    mysum = 0
    detect_cancellation = False

    async def mytask() -> None:
        nonlocal mysum, detect_cancellation
        try:
            async with aiotools.aclosing(myiter()) as agen:
                async for x in agen:
                    mysum += x
        except asyncio.CancelledError:
            detect_cancellation = True

    t = asyncio.create_task(mytask())
    await asyncio.sleep(0.3)
    t.cancel()
    await t

    assert mysum == 11
    assert finalized
    assert detect_cancellation



"""
Based on :doc:`aiotools.context`, this module provides an automated lifecycle
management for multi-process servers with explicit initialization steps and
graceful shutdown steps.

.. function:: server(func)

    A decorator wrapper for :class:`AsyncServerContextManager`.

    Usage example:

    .. code:: python

       @aiotools.server
       async def myserver(loop, pidx, args):
           await do_init(args)
           stop_sig = yield
           if stop_sig == signal.SIGINT:
               await do_graceful_shutdown()
           else:
               await do_forced_shutdown()

       aiotools.start_server(myserver, ...)
"""

import asyncio
import functools
import inspect
import logging
import multiprocessing as mp
import multiprocessing.connection as mpconn
import os
import signal
import struct
import sys
import threading
from collections.abc import (
    AsyncGenerator,
    Callable,
    Collection,
    Generator,
    Mapping,
    Sequence,
)
from contextlib import AbstractContextManager, ContextDecorator
from contextvars import ContextVar
from types import TracebackType
from typing import Any, ParamSpec, TypeVar

from .compat import all_tasks, current_task, get_running_loop
from .context import AbstractAsyncContextManager
from .fork import AbstractChildProcess, MPContext, _has_pidfd, afork

__all__ = (
    "main_context",
    "server_context",
    "start_server",
    "process_index",
    "AsyncServerContextManager",
    "ServerMainContextManager",
    "InterruptedBySignal",
)

log = logging.getLogger(__name__)

process_index: ContextVar[int] = ContextVar("process_index")

TYield = TypeVar("TYield")
PArgs = ParamSpec("PArgs")


class InterruptedBySignal(BaseException):
    """
    A new :class:`BaseException` that represents interruption by an arbitrary UNIX
    signal.

    Since this is a :class:`BaseException` instead of :class:`Exception`, it behaves
    like :class:`KeyboardInterrupt` and :class:`SystemExit` exceptions (i.e.,
    bypassing except clauses catching the :class:`Exception` type only)

    The first argument of this exception is the signal number received.
    """

    pass


class AsyncServerContextManager(AbstractAsyncContextManager[TYield]):
    """
    A modified version of :func:`contextlib.asynccontextmanager`.

    The implementation detail is mostly taken from the ``contextlib`` standard
    library, with a minor change to inject ``self.yield_return`` into the wrapped
    async generator.
    """

    yield_return: signal.Signals = signal.SIGTERM

    def __init__(
        self,
        func: Callable[..., AsyncGenerator[TYield, signal.Signals]],
        args: Sequence[Any],
        kwargs: Mapping[str, Any],
    ) -> None:
        if not inspect.isasyncgenfunction(func):
            raise RuntimeError("Context manager function must be an async-generator")
        self._agen: AsyncGenerator[TYield, signal.Signals] = func(*args, **kwargs)
        self.func: Callable[..., AsyncGenerator[TYield, signal.Signals]] = func
        self.args: Sequence[Any] = args
        self.kwargs: Mapping[str, Any] = kwargs
        self.yield_return: signal.Signals = signal.SIGTERM

    async def __aenter__(self) -> TYield:
        try:
            return await self._agen.__anext__()
        except StopAsyncIteration:
            raise RuntimeError("async-generator didn't yield") from None

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
        /,
    ) -> bool | None:
        if exc_type is None:
            try:
                # Here is the modified part.
                await self._agen.asend(self.yield_return)
            except StopAsyncIteration:
                return None
            else:
                raise RuntimeError("async-generator didn't stop") from None
        else:
            if exc_value is None:
                exc_value = exc_type()
            try:
                await self._agen.athrow(exc_type, exc_value, traceback)
                raise RuntimeError("async-generator didn't stop after athrow()")
            except StopAsyncIteration as exc_new_value:
                return exc_new_value is not exc_value
            except RuntimeError as exc_new_value:
                if exc_new_value is exc_value:
                    return False
                if isinstance(exc_value, (StopIteration, StopAsyncIteration)):
                    if exc_new_value.__cause__ is exc_value:
                        return False
                raise
            except (BaseException, asyncio.CancelledError) as exc:
                if exc is not exc_value:
                    raise
        return None


class ServerMainContextManager(AbstractContextManager[TYield], ContextDecorator):
    """
    A modified version of :func:`contextlib.contextmanager`.

    The implementation detail is mostly taken from the ``contextlib`` standard
    library, with a minor change to inject ``self.yield_return`` into the wrapped
    generator.
    """

    def __init__(
        self,
        func: Callable[..., Generator[TYield, signal.Signals, None]],
        args: Sequence[Any],
        kwargs: Mapping[str, Any],
    ) -> None:
        self.gen: Generator[TYield, signal.Signals, None] = func(*args, **kwargs)
        self.func: Callable[..., Generator[TYield, signal.Signals, None]] = func
        self.args: Sequence[Any] = args
        self.kwargs: Mapping[str, Any] = kwargs
        self.yield_return: signal.Signals = signal.SIGTERM

    def __enter__(self) -> TYield:
        del self.args, self.kwargs, self.func
        try:
            return next(self.gen)
        except StopIteration:
            raise RuntimeError("generator didn't yield") from None

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
        /,
    ) -> bool | None:
        if exc_type is None:
            try:
                self.gen.send(self.yield_return)
            except StopIteration:
                return False
            else:
                raise RuntimeError("generator didn't stop")
        else:
            if exc_value is None:
                exc_value = exc_type()
            try:
                self.gen.throw(exc_type, exc_value, traceback)
            except StopIteration as exc:
                return exc is not exc_value
            except RuntimeError as exc:
                if exc is exc_value:
                    return False
                if exc_type is StopIteration and exc.__cause__ is exc_value:
                    return False
                raise
            except:  # noqa
                if sys.exc_info()[1] is exc_value:
                    return False
                raise
            raise RuntimeError("generator didn't stop after throw()")
        return None


# This is a dirty hack to implement "module callable".
# NOTE: only works in Python 3.5 or higher.


def server_context(
    func: Callable[
        [asyncio.AbstractEventLoop, int, Sequence[Any]],
        AsyncGenerator[None, signal.Signals],
    ],
) -> Callable[..., AsyncServerContextManager[None]]:
    @functools.wraps(func)
    def helper(*args: Any, **kwargs: Any) -> AsyncServerContextManager[None]:
        return AsyncServerContextManager(func, args, kwargs)

    return helper


class _ServerModule(sys.modules[__name__].__class__):  # type: ignore
    def __call__(
        self,
        func: Callable[
            [asyncio.AbstractEventLoop, int, Sequence[Any]],
            AsyncGenerator[None, signal.Signals],
        ],
    ) -> Callable[..., AsyncServerContextManager[None]]:
        return server_context(func)


sys.modules[__name__].__class__ = _ServerModule


def main_context(
    func: Callable[[], Generator[TYield, signal.Signals, None]],
) -> Callable[[], ServerMainContextManager[TYield]]:
    """
    A decorator wrapper for :class:`ServerMainContextManager`

    Usage example:

    .. code:: python

       @aiotools.main_context
       def mymain() -> Generator[TServerArgs, signal.Signals]:
           server_args = do_init()
           stop_sig = yield server_args
           if stop_sig == signal.SIGINT:
               do_graceful_shutdown()
           else:
               do_forced_shutdown()

       aiotools.start_server(..., main_ctxmgr=mymain, ...)
    """

    @functools.wraps(func)
    def helper(*args: Any, **kwargs: Any) -> ServerMainContextManager[TYield]:
        return ServerMainContextManager(func, args, kwargs)

    return helper


def setup_child_watcher(loop: asyncio.AbstractEventLoop) -> None:
    if sys.version_info < (3, 12, 0):
        # see python/cpython#94597 (issue) and python/cpython#98215 (pr)
        try:
            watcher_cls = getattr(asyncio, "PidfdChildWatcher", None)
            if _has_pidfd and watcher_cls:
                watcher = watcher_cls()
                asyncio.set_child_watcher(watcher)
            else:
                # Just get the default child watcher.
                watcher = asyncio.get_child_watcher()
            if not watcher.is_active():
                watcher.attach_loop(loop)
        except NotImplementedError:
            pass  # for uvloop


async def cancel_all_tasks() -> None:
    loop = get_running_loop()
    cancelled_tasks: list[asyncio.Task[Any]] = []
    for task in all_tasks():
        if not task.done() and task is not current_task():
            task.cancel()
            cancelled_tasks.append(task)
    await asyncio.gather(*cancelled_tasks, return_exceptions=True)
    for task in cancelled_tasks:
        if task.cancelled():
            continue
        if task.exception() is not None:
            loop.call_exception_handler({
                "message": "unhandled exception during loop shutdown",
                "exception": task.exception(),
                "task": task,
            })


def _get_default_stop_signal(
    stop_signals: Collection[signal.Signals],
) -> signal.Signals:
    if signal.SIGTERM in stop_signals:  # prefer SIGTERM if included
        return signal.SIGTERM
    return sorted(stop_signals)[0]


def _worker_main(
    worker_actxmgr: Callable[
        [asyncio.AbstractEventLoop, int, Sequence[Any]],
        AsyncServerContextManager[Any],
    ],
    stop_signals: Collection[signal.Signals],
    intr_write_pipe: mpconn.Connection,
    proc_idx: int,
    run_to_completion: bool,
    args: Sequence[Any],
    *,
    prestart_hook: Callable[[int], None] | None = None,
) -> int:
    process_index.set(proc_idx)
    if prestart_hook:
        prestart_hook(proc_idx)
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    setup_child_watcher(loop)
    interrupted = asyncio.Event()
    ctx = worker_actxmgr(loop, proc_idx, args)
    forever_future: asyncio.Future[None] = loop.create_future()

    def handle_stop_signal(signum: signal.Signals) -> None:
        if interrupted.is_set():
            pass
        else:
            interrupted.set()
            ctx.yield_return = signum
            forever_future.cancel()

    for signum in stop_signals:
        loop.add_signal_handler(
            signum,
            functools.partial(handle_stop_signal, signum),
        )
    # Allow the worker to be interrupted during initialization
    # (in case of initialization failures in other workers)
    signal.pthread_sigmask(signal.SIG_UNBLOCK, stop_signals)

    async def _wrapped_worker() -> None:
        err_ctx = "enter"
        try:
            async with ctx:
                err_ctx = "body"
                try:
                    if not run_to_completion:
                        await forever_future
                    else:
                        ctx.yield_return = _get_default_stop_signal(stop_signals)
                except asyncio.CancelledError:
                    pass
                finally:
                    err_ctx = "exit"
        except Exception:
            if err_ctx != "body":
                err_ctx_str = "initialization" if err_ctx == "enter" else "shutdown"
                log.exception(
                    f"Worker {proc_idx}: Error during context manager {err_ctx_str}"
                )
                try:
                    intr_write_pipe.send_bytes(struct.pack("i", proc_idx))
                except BrokenPipeError:
                    pass
            raise
        finally:
            intr_write_pipe.close()

    try:
        loop.run_until_complete(_wrapped_worker())
    finally:
        try:
            loop.run_until_complete(cancel_all_tasks())
            loop.run_until_complete(loop.shutdown_asyncgens())
            try:
                loop.run_until_complete(loop.shutdown_default_executor())
            except (AttributeError, NotImplementedError):  # for uvloop
                pass
        finally:
            loop.close()
        return 0


def _extra_main(
    extra_func: Callable[[threading.Event | None, int, Sequence[Any]], None],
    stop_signals: Collection[signal.Signals],
    proc_idx: int,
    args: Sequence[Any],
    *,
    prestart_hook: Callable[[int], None] | None = None,
) -> int:
    process_index.set(proc_idx)
    if prestart_hook:
        prestart_hook(proc_idx)
    interrupted = threading.Event()

    # Since signals only work for the main thread in Python,
    # extra processes in use_threading=True mode should check
    # the intr_event by themselves (probably in their loops).

    def raise_stop(signum: int, frame: Any) -> None:
        if interrupted.is_set():
            pass
        else:
            interrupted.set()
            if signum == signal.SIGINT:
                raise KeyboardInterrupt
            elif signum == signal.SIGTERM:
                raise SystemExit
            else:
                raise InterruptedBySignal(signum)

    # restore signal handler.
    for signum in stop_signals:
        signal.signal(signum, raise_stop)
    signal.pthread_sigmask(signal.SIG_UNBLOCK, stop_signals)
    intr_event: threading.Event | None = None

    try:
        if not interrupted.is_set():
            extra_func(intr_event, proc_idx, args)
    except (SystemExit, KeyboardInterrupt, InterruptedBySignal):
        log.warning(f"extra_proc[{proc_idx}] did not handle stop signals.")
    finally:
        # same as in _worker_main()
        signal.pthread_sigmask(signal.SIG_BLOCK, stop_signals)
        return 0


def start_server(
    worker_actxmgr: Callable[
        [asyncio.AbstractEventLoop, int, Sequence[Any]], AsyncServerContextManager[Any]
    ],
    main_ctxmgr: Callable[[], ServerMainContextManager[Any]] | None = None,
    extra_procs: Collection[
        Callable[[threading.Event | None, int, Sequence[Any]], None]
    ] = tuple(),
    stop_signals: Collection[signal.Signals] = (signal.SIGINT, signal.SIGTERM),
    num_workers: int = 1,
    args: Sequence[Any] = tuple(),
    *,
    wait_timeout: float | None = None,
    mp_context: MPContext | None = None,
    prestart_hook: Callable[[int], None] | None = None,
    ignore_child_interrupts: bool = False,
    run_to_completion: bool = False,
) -> None:
    """
    Starts a multi-process server where each process has their own individual
    asyncio event loop.  Their lifecycles are automantically managed -- if the
    main program receives one of the signals specified in ``stop_signals`` it
    will initiate the shutdown routines on each worker that stops the event
    loop gracefully.

    Args:
        worker_actxmgr:
            An asynchronous context manager that dicates the
            initialization and shutdown steps of each worker.
            It should accept the following three arguments:

            * **loop**: the asyncio event loop created and set
              by aiotools
            * **pidx**: the 0-based index of the worker
              (use this for per-worker logging)
            * **args**: a concatenated tuple of values yielded by
              **main_ctxmgr** and the user-defined arguments in
              **args**.

            aiotools automatically installs an interruption handler
            that calls ``loop.stop()`` to the given event loop,
            regardless of using either threading or
            multiprocessing.

        main_ctxmgr:
            An optional context manager that performs global
            initialization and shutdown steps of the whole program.
            It may yield one or more values to be passed to worker
            processes along with **args** passed to this function.
            There is no arguments passed to those functions since
            you can directly access ``sys.argv`` to parse command
            line arguments and/or read user configurations.

        extra_procs:
            An iterable of functions that consist of extra processes
            whose lifecycles are synchronized with other workers.
            They should set up their own signal handlers.

            It should accept the following three arguments:

            * **intr_event**: Always ``None``, kept for legacy
            * **pidx**: same to **worker_actxmgr** argument
            * **args**: same to **worker_actxmgr** argument

        stop_signals:
            A list of UNIX signals that the main program to
            recognize as termination signals.

        num_workers: The number of children workers.

        args:
            The user-defined arguments passed to workers and extra
            processes.  If **main_ctxmgr** yields one or more values,
            they are *prepended* to this user arguments when passed to
            workers and extra processes.

        wait_timeout:
            The timeout in seconds before forcibly killing all
            remaining child processes after sending initial stop
            signals.

        mp_context:
            The multiprocessing context to use for creating child
            processes.  If not specified, the default context is
            used.

        prestart_hook:
            A function to be called once before creating the
            event loop in the children.  The function should
            accept an int argument representing the process index.
            You may use this hook to initialize the event loop
            with custom settings (e.g., applying ``uvloop``)
            or to perform any other necessary setup before the
            main function is called.

        ignore_child_interrupts:
            By default, any unhandled exceptions in the
            child functions are translated as active
            SIGINT to the main and all other worker processes,
            meaning a full shutdown.
            This flag makes the main process to ignore them,
            which is useful to gather all worker's results
            even when some of them raises unhandled exceptions.

        run_to_completion:
            If True, the main/worker processes will NOT wait forever
            until interrupted but *immediately exit* when the main
            functions complete.  This flag implies **ignore_child_interrupts**.

    Returns:
        None

    .. versionchanged:: 0.3.2

       The name of argument **num_proc** is changed to **num_workers**.
       Even if **num_workers** is 1, a child is created instead of
       doing everything at the main thread.

    .. versionadded:: 0.3.2

       The argument ``extra_procs`` and ``main_ctxmgr``.

    .. versionadded:: 0.4.0

       Now supports use of threading instead of multiprocessing via
       **use_threading** option.

    .. versionchanged:: 0.8.0

       Now **worker_actxmgr** must be an instance of
       :class:`AsyncServerContextManager` or async generators decorated by
       ``@aiotools.server``.

       Now **main_ctxmgr** must be an instance of :class:`ServerMainContextManager`
       or plain generators decorated by ``@aiotools.main``.

       The usage is same to asynchronous context managers, but optionally you can
       distinguish the received stop signal by retrieving the return value of the
       ``yield`` statement.

       In **extra_procs** in non-threaded mode, stop signals are converted into
       either one of :class:`KeyboardInterrupt`, :class:`SystemExit`, or
       :class:`InterruptedBySignal` exception.

    .. versionadded:: 0.8.4

       **start_method** argument can be set to change the subprocess spawning
       implementation.

    .. deprecated:: 1.2.0

       The **start_method** and **use_threading** arguments, in favor of our new
       :func:`afork()` function which provides better synchronization and pid-fd
       support.

    .. versionchanged:: 1.2.0

       The **extra_procs** will be always separate processes since **use_threading**
       is deprecated and thus **intr_event** arguments are now always ``None``.

    .. versionadded:: 1.5.5

       The **wait_timeout** argument.

    .. versionadded:: 1.9.0

        The **mp_context**, **prestart_hook**, **ignore_child_interrupts**,
        and **run_to_completion** arguments.
    """

    @main_context
    def noop_main_ctxmgr() -> Generator[None, signal.Signals, None]:
        yield

    assert stop_signals

    if hasattr(asyncio, "get_running_loop"):
        # only for Python 3.7+
        try:
            asyncio.get_running_loop()
        except RuntimeError:
            # there should be none.
            pass
        else:
            raise RuntimeError(
                "aiotools.start_server() cannot be called inside a running event loop."
            )

    if main_ctxmgr is None:
        main_ctxmgr = noop_main_ctxmgr

    children: list[AbstractChildProcess] = []
    sigblock_mask: frozenset[signal.Signals] = frozenset(stop_signals)

    main_ctx: ServerMainContextManager[Any] = main_ctxmgr()

    # temporarily block signals and register signal handlers to main_loop
    signal.pthread_sigmask(signal.SIG_BLOCK, sigblock_mask)

    main_loop = asyncio.new_event_loop()
    asyncio.set_event_loop(main_loop)
    main_future: asyncio.Future[None] = main_loop.create_future()

    # to make subprocess working in child threads
    setup_child_watcher(main_loop)

    # Build a main-to-worker interrupt channel using signals
    def handle_stop_signal(signum: signal.Signals) -> None:
        main_ctx.yield_return = signum
        for child in children:
            child.send_signal(signum)
        # instead of main_loop.stop(), we use a future here
        # NOT to interrupt the cleanup routines in
        # an arbitrary timing upon async child failures.
        main_future.cancel()

    for signum in stop_signals:
        main_loop.add_signal_handler(
            signum,
            functools.partial(handle_stop_signal, signum),
        )

    # Build a reliable worker-to-main interrupt channel using a pipe.
    # This channel is used when the worker main functions raise an unhandled exception,
    # so that the main program can be interrupted immediately.
    def handle_child_interrupt(read_pipe: mpconn.Connection) -> None:
        try:
            child_idx: int = struct.unpack("i", read_pipe.recv_bytes(4))[0]
        except EOFError:
            # read_pipe is already closed
            return
        if not ignore_child_interrupts and not run_to_completion:
            # self-interrupt to initiate the main-to-worker interrupts
            log.debug(f"Child {child_idx} has interrupted the main program.")
            signal.pthread_sigmask(signal.SIG_UNBLOCK, {signal.SIGINT})
            os.kill(0, signal.SIGINT)

    read_pipe, write_pipe = mp.Pipe()
    main_loop.add_reader(read_pipe.fileno(), handle_child_interrupt, read_pipe)

    # start
    try:
        with main_ctx as main_args:
            # retrieve args generated by the user-defined main
            main_args_tuple: tuple[Any, ...]
            if main_args is None:
                main_args_tuple = tuple()
            elif not isinstance(main_args, tuple):
                main_args_tuple = (main_args,)
            else:
                main_args_tuple = main_args

            # spawn managed async workers
            for i in range(num_workers):
                try:
                    p = main_loop.run_until_complete(
                        afork(
                            functools.partial(
                                _worker_main,
                                worker_actxmgr,
                                stop_signals,
                                write_pipe,
                                i,
                                run_to_completion,
                                (*main_args_tuple, *args),
                                prestart_hook=prestart_hook,
                            ),
                            mp_context=mp_context,
                        )
                    )
                except RuntimeError as e:
                    if "loop stopped" in e.args[0]:
                        log.warning(
                            "skipping spawning of child[worker]:%d due to "
                            "async failure(s) of other child process",
                            i,
                        )
                        continue
                    raise
                children.append(p)

            # spawn extra workers
            for i, f in enumerate(extra_procs):
                try:
                    p = main_loop.run_until_complete(
                        afork(
                            functools.partial(
                                _extra_main,
                                f,
                                stop_signals,
                                num_workers + i,
                                (*main_args_tuple, *args),
                                prestart_hook=prestart_hook,
                            ),
                            mp_context=mp_context,
                        )
                    )
                except RuntimeError as e:
                    if "loop stopped" in e.args[0]:
                        log.warning(
                            "skipping spawning of child[extra]:%d due to "
                            "async failure(s) of other child process",
                            num_workers + i,
                        )
                        continue
                    raise
                children.append(p)

            write_pipe.close()

            # unblock the stop signals for user/external interrupts.
            signal.pthread_sigmask(signal.SIG_UNBLOCK, sigblock_mask)

            # run!
            try:
                if not run_to_completion:
                    main_loop.run_until_complete(main_future)
                else:
                    main_ctx.yield_return = _get_default_stop_signal(stop_signals)
            except asyncio.CancelledError:
                pass
            finally:
                # If interrupted or complete, wait for workers to finish.
                try:
                    worker_results: list[int | BaseException] = (
                        main_loop.run_until_complete(
                            asyncio.wait_for(
                                asyncio.gather(
                                    *[child.wait() for child in children],
                                    return_exceptions=True,
                                ),
                                wait_timeout,
                            )
                        )
                    )
                    for child, result in zip(children, worker_results):
                        if isinstance(result, Exception):
                            log.error(
                                "Waiting for a child process [%d] has failed by an error.",
                                child.pid,
                                exc_info=result,
                            )
                except asyncio.TimeoutError:
                    log.warning(
                        "Timeout during waiting for child processes; killing all",
                    )
                    for child in children:
                        child.send_signal(signal.SIGKILL)
    finally:
        main_loop.remove_reader(read_pipe.fileno())
        read_pipe.close()
        try:
            main_loop.run_until_complete(cancel_all_tasks())
            main_loop.run_until_complete(main_loop.shutdown_asyncgens())
            try:
                main_loop.run_until_complete(main_loop.shutdown_default_executor())
            except (AttributeError, NotImplementedError):  # for uvloop
                pass
        finally:
            main_loop.close()
            asyncio.set_event_loop(None)



from __future__ import annotations

import asyncio
import warnings
from typing import Any

__all__ = (
    "all_tasks",
    "get_running_loop",
    "current_task",
    "set_task_name",
)

get_running_loop = asyncio.get_running_loop
all_tasks = asyncio.all_tasks


def current_task() -> asyncio.Task[Any]:
    """
    A simple null-guarded version of :func:`asyncio.current_task()` wrapper
    to simplify type handling in the caller side.
    """
    task = asyncio.current_task()
    if task is None:
        raise RuntimeError(
            "The current stack does not belong to any running asyncio task."
        )
    return task


def set_task_name(task: asyncio.Task[Any], name: str | None) -> None:
    # This compatibility function had been in asyncio.tasks until Python 3.12,
    # but removed since Python 3.13.
    if name is not None:
        try:
            set_name = task.set_name
        except AttributeError:
            warnings.warn(
                "Task name customization may not be available in 3rd-party event loops before Python 3.13",
                DeprecationWarning,
                stacklevel=2,
            )
        else:
            set_name(name)



from __future__ import annotations

import enum
from builtins import aiter as _builtin_aiter
from collections.abc import AsyncIterable, AsyncIterator, Awaitable, Callable
from typing import Final, TypeVar, cast, overload

__all__ = ("aiter",)

_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)


class _Sentinel(enum.Enum):
    MISSING = 0


MISSING: Final = _Sentinel.MISSING


@overload
def aiter(source: AsyncIterable[_T], /) -> AsyncIterator[_T]: ...


@overload
def aiter(
    func: Callable[[], Awaitable[_T]], sentinel: _T | _Sentinel, /
) -> AsyncIterator[_T]: ...


def aiter(
    arg: AsyncIterable[_T] | Callable[[], Awaitable[_T]],
    sentinel: _T | _Sentinel = MISSING,
) -> AsyncIterator[_T]:
    """
    Analogous to the builtin :func:`iter()`.

    As of Python 3.10, this is a small extension accepting an explicit sentinel
    while the base implementation without sentinel uses the builtin :func:`aiter()`.
    """

    # The forced typecasting here looks ugly, but it is an inevitable choice:
    # - runtime checks using functools.singledispatch cannot evaluate type annotations with generics.
    # - runtime checks using the pattern matching syntax has the same limitation.
    # -

    if sentinel is MISSING:
        # Fast-path using the builtin directly.
        return _builtin_aiter(cast(AsyncIterable[_T], arg))

    func = cast(Callable[[], Awaitable[_T]], arg)

    # The "yield" statement changes the type of function to generator,
    # so we need to separate the type using an inner function.
    async def _gen() -> AsyncIterator[_T]:
        while True:
            item = await func()
            if item == sentinel:
                break
            yield item

    return _gen()



import asyncio
import signal
from collections.abc import AsyncGenerator, Sequence
from typing import Any

import aiotools


async def echo(reader, writer):
    data = await reader.read(100)
    writer.write(data)
    await writer.drain()
    writer.close()


@aiotools.server_context
async def worker_main(
    loop: asyncio.AbstractEventLoop,
    pidx: int,
    args: Sequence[Any],
) -> AsyncGenerator[None, signal.Signals]:
    # Create a listening socket with SO_REUSEPORT option so that each worker
    # process can share the same listening port and the kernel balances
    # incoming connections across multiple worker processes.
    server = await asyncio.start_server(
        echo, "0.0.0.0", 8888, reuse_port=True, loop=loop
    )
    print(f"[{pidx}] started")

    sig = None
    try:
        sig = yield  # wait until terminated
    finally:
        server.close()
        await server.wait_closed()
        print(f"[{pidx}] terminated with {sig=}")


if __name__ == "__main__":
    # Run the above server using 4 worker processes.
    aiotools.start_server(worker_main, num_workers=4)



"""
A set of common utilities for legacy Pythons
"""

from __future__ import annotations

import asyncio
from typing import TypeVar

from ..types import CoroutineLike

_has_task_name: bool = hasattr(asyncio.Task, "get_name")
T = TypeVar("T")


def create_task_with_name(
    coro: CoroutineLike[T], *, name: str | None = None
) -> asyncio.Task[T]:
    loop = asyncio.get_running_loop()
    if _has_task_name and name:
        child_task = loop.create_task(coro, name=name)
    else:
        child_task = loop.create_task(coro)
    return child_task



"""
Provides a Golang-like ``defer()`` API using decorators, which allows grouping
resource initialization and cleanup in one place without extra indentations.

Example:

.. code-block:: python3

   async def init(x):
       ...

   async def cleanup(x):
       ...

   @aiotools.adefer
   async def do(defer):  # <-- be aware of defer argument!
       x = SomeResource()
       await init(x)
       defer(cleanup(x))
       ...
       ...

This is equivalent to:

.. code-block:: python3

   async def do():
       x = SomeResource()
       await init(x)
       try:
           ...
           ...
       finally:
           await cleanup(x)

Note that :class:`aiotools.context.AsyncContextGroup` or
:class:`contextlib.AsyncExitStack` serves well for the same purpose, but for simple
cleanups, this defer API makes your codes simple because it steps aside the main
execution context without extra indentations.

.. warning::

   Any exception in the deferred functions is raised transparently, and may block
   execution of the remaining deferred functions.
   This behavior may be changed in the future versions, though.
"""

from __future__ import annotations

import functools
import inspect
from collections import deque
from collections.abc import Awaitable, Callable
from typing import Any, Concatenate, ParamSpec, TypeAlias, TypeVar

P = ParamSpec("P")
R = TypeVar("R")

__all__ = (
    "defer",
    "adefer",
    "DeferFunc",
    "AsyncDeferrable",
    "AsyncDeferFunc",
)

DeferFunc: TypeAlias = Callable[[Callable[[], None]], None]
AsyncDeferrable: TypeAlias = (
    Callable[[], Awaitable[Any]] | Callable[[], None] | Awaitable[None]
)
AsyncDeferFunc: TypeAlias = Callable[[AsyncDeferrable], None]


def defer(func: Callable[Concatenate[DeferFunc, P], R]) -> Callable[P, R]:
    """
    A synchronous version of the defer API.
    It can only defer normal functions.
    """
    assert not inspect.iscoroutinefunction(func), (
        "the decorated function must not be async"
    )

    @functools.wraps(func)
    def _wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
        deferreds: deque[Callable[[], None]] = deque()

        # the registration function passed to the original function
        def defer(f: Callable[[], None]) -> None:
            assert not inspect.iscoroutinefunction(f), (
                "the deferred function must not be async"
            )
            assert not inspect.iscoroutine(f), (
                "the deferred object must not be a coroutine"
            )
            deferreds.append(f)

        try:
            return func(defer, *args, **kwargs)
        finally:
            while deferreds:
                f = deferreds.pop()
                f()

    return _wrapped


def adefer(
    func: Callable[Concatenate[AsyncDeferFunc, P], Awaitable[R]],
) -> Callable[P, Awaitable[R]]:
    """
    An asynchronous version of the defer API.
    It can defer coroutine functions, coroutines, and normal functions.
    """
    assert inspect.iscoroutinefunction(func), "the decorated function must be async"

    @functools.wraps(func)
    async def _wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
        deferreds: deque[AsyncDeferrable] = deque()

        # the registration function passed to the original function
        def defer(f: AsyncDeferrable) -> None:
            deferreds.append(f)

        try:
            return await func(defer, *args, **kwargs)
        finally:
            while deferreds:
                f = deferreds.pop()
                if inspect.iscoroutinefunction(f):
                    await f()
                elif inspect.iscoroutine(f):
                    await f
                else:
                    f()  # type: ignore[operator]

    return _wrapped



import asyncio
import logging
import os
import signal
import threading
from collections.abc import AsyncGenerator, Sequence
from typing import Any

import zmq
import zmq.asyncio

import aiotools

num_workers = 4

log_init_states: set[str] = set()


def get_logger(name: str, pid: int) -> logging.Logger:
    log = logging.getLogger(name)
    # initialize only once for each logger identified by the name
    if name not in log_init_states:
        sh = logging.StreamHandler()
        fmt = logging.Formatter(
            f"%(relativeCreated).3f %(name)s[{pid}] %(levelname)s: %(message)s"
        )
        sh.setFormatter(fmt)
        log.addHandler(sh)
        log.propagate = False
        log.setLevel(logging.INFO)
        log_init_states.add(name)
    return log


def router_main(
    intr_event: threading.Event | None,
    pidx: int,
    args: Sequence[Any],
) -> None:
    log = get_logger("examples.zmqserver.extra", pidx)
    zctx = zmq.Context()
    zctx.linger = 0
    in_sock = zctx.socket(zmq.PULL)
    in_sock.bind("tcp://*:5033")
    out_sock = zctx.socket(zmq.PUSH)
    out_sock.bind("ipc://example-events")
    try:
        log.info("router proxy started")
        zmq.proxy(in_sock, out_sock)
    except KeyboardInterrupt:
        pass
    except Exception:
        log.exception("unexpected error")
    finally:
        for _ in range(num_workers):
            out_sock.send(b"")  # sentinel
        log.info("router proxy terminated")
        in_sock.close()
        out_sock.close()
        zctx.term()
        os.unlink("example-events")


@aiotools.server_context
async def worker_main(
    loop: asyncio.AbstractEventLoop,
    pidx: int,
    args: Sequence[Any],
) -> AsyncGenerator[None, signal.Signals]:
    log = get_logger("examples.zmqserver.worker", pidx)
    zctx = zmq.asyncio.Context()
    router = zctx.socket(zmq.PULL)
    router.connect("ipc://example-events")

    async def process_incoming(router):
        while True:
            data = await router.recv()
            if not data:
                return
            log.info(data)

    task = loop.create_task(process_incoming(router))
    log.info("started")

    sig = None
    try:
        sig = yield
    finally:
        await task
        router.close()
        zctx.term()
        log.info(f"terminated with {sig=}")


if __name__ == "__main__":
    # This example must be run with multiprocessing.
    aiotools.start_server(
        worker_main,
        num_workers=num_workers,
        extra_procs=[router_main],
    )



"""
Provides an implementation of asynchronous context manager and its applications.

.. note::

   The async context managers in this module are transparent aliases to
   ``contextlib.asynccontextmanager`` of the standard library in Python 3.7
   and later.
"""

import asyncio
import contextlib
from collections.abc import Iterable
from contextvars import ContextVar
from typing import Any, Generic, TypeVar

from .types import AsyncClosable, OptExcInfo

__all__ = [
    "resetting",
    "AsyncContextManager",
    "async_ctx_manager",
    "actxmgr",
    "aclosing",
    "closing_async",
    "AsyncContextGroup",
    "actxgroup",
    "AsyncExitStack",
]


T = TypeVar("T")
T_AsyncClosable = TypeVar("T_AsyncClosable", bound=AsyncClosable)

AbstractAsyncContextManager = contextlib.AbstractAsyncContextManager
AsyncContextManager = contextlib._AsyncGeneratorContextManager
AsyncExitStack = contextlib.AsyncExitStack
async_ctx_manager = contextlib.asynccontextmanager
aclosing = contextlib.aclosing


class resetting(Generic[T]):
    """
    An extra context manager to auto-reset the given context variable.
    It supports both the standard contextmanager protocol and the
    async-contextmanager protocol.

    .. versionadded:: 1.8.0
    """

    def __init__(self, ctxvar: ContextVar[T], value: T) -> None:
        self._ctxvar = ctxvar
        self._value = value

    def __enter__(self) -> None:
        self._token = self._ctxvar.set(self._value)

    async def __aenter__(self) -> None:
        self._token = self._ctxvar.set(self._value)

    def __exit__(self, *exc_info: OptExcInfo) -> bool | None:
        self._ctxvar.reset(self._token)
        return None

    async def __aexit__(self, *exc_info: OptExcInfo) -> bool | None:
        self._ctxvar.reset(self._token)
        return None


class closing_async(Generic[T_AsyncClosable]):
    """
    An analogy to :func:`contextlib.closing` for objects defining the ``close()``
    method as an async function.

    .. versionadded:: 1.5.6
    """

    def __init__(self, thing: T_AsyncClosable) -> None:
        self.thing = thing

    async def __aenter__(self) -> T_AsyncClosable:
        return self.thing

    async def __aexit__(self, *exc_info: OptExcInfo) -> bool | None:
        await self.thing.close()
        return None


class AsyncContextGroup:
    """
    Merges a group of context managers into a single context manager.
    Internally it uses :func:`asyncio.gather()` to execute them with overlapping,
    to reduce the execution time via asynchrony.

    Upon entering, you can get values produced by the entering steps from
    the passed context managers (those ``yield``-ed) using an ``as`` clause of
    the ``async with``
    statement.

    After exits, you can check if the context managers have finished
    successfully by ensuring that the return values of ``exit_states()`` method
    are ``None``.

    .. note::

       You cannot return values in context managers because they are
       generators.

    If an exception is raised before the ``yield`` statement of an async
    context manager, it is stored at the corresponding manager index in the
    as-clause variable.  Similarly, if an exception is raised after the
    ``yield`` statement of an async context manager, it is stored at the
    corresponding manager index in the ``exit_states()`` return value.

    Any exceptions in a specific context manager does not interrupt others;
    this semantic is same to ``asyncio.gather()``'s when
    ``return_exceptions=True``.  This means that, it is user's responsibility
    to check if the returned context values are exceptions or the intended ones
    inside the context body after entering.

    :param context_managers: An iterable of async context managers.
                             If this is ``None``, you may add async context
                             managers one by one using the :meth:`~.add`
                             method.

    """

    def __init__(
        self, context_managers: Iterable[AbstractAsyncContextManager[Any]] | None = None
    ) -> None:
        self._cm: list[AbstractAsyncContextManager[Any]] = (
            list(context_managers) if context_managers else []
        )
        self._cm_yields: list[Any | BaseException] = []
        self._cm_exits: list[bool | None | BaseException] = []

    def add(self, cm: AbstractAsyncContextManager[Any]) -> None:
        self._cm.append(cm)

    async def __aenter__(self) -> list[Any | BaseException]:
        # Exceptions in context managers are stored into _cm_yields list.
        # NOTE: There is no way to "skip" the context body even if the entering
        #       process fails.
        self._cm_yields[:] = await asyncio.gather(
            *(e.__aenter__() for e in self._cm), return_exceptions=True
        )
        return self._cm_yields

    async def __aexit__(self, *exc_info: OptExcInfo) -> None:
        # Clear references to context variables.
        self._cm_yields.clear()
        # Exceptions are stored into _cm_exits list.
        # The type-ignore comment is required to embrace the difference between T_T:
        # types.OptExcInfo                                      = tuple[type[BaseException], BaseException, TracebackType] | tuple[None, None, None]
        # args@contextlib.AbstractAsyncContextManager.__aexit__ = tuple[type[BaseException] | None, BaseException | None, TracebackType | None]
        self._cm_exits[:] = await asyncio.gather(
            *(e.__aexit__(*exc_info) for e in self._cm),  # type: ignore[arg-type]
            return_exceptions=True,
        )

    def exit_states(self) -> list[bool | None | BaseException]:
        return self._cm_exits


# Shorter aliases
actxmgr = async_ctx_manager
actxgroup = AsyncContextGroup



from collections.abc import Sequence
from types import TracebackType
from typing import Protocol


class AsyncExceptionHandler(Protocol):
    """
    A shorthand for an async exception handler type.
    This is always called under exception context where
    :func:`sys.exc_info()` is available.
    """

    async def __call__(
        self,
        exc_type: type[BaseException],
        exc_obj: BaseException,
        exc_tb: TracebackType,
    ) -> None: ...


class MultiError(BaseExceptionGroup):
    __errors__: Sequence[BaseException]

    def __init__(self, msg: str, errors: Sequence[BaseException], /) -> None:
        super().__init__(msg, errors)
        self.__errors__ = errors

    def get_error_types(self) -> set[type[BaseException]]:
        return {type(e) for e in self.exceptions}


class TaskGroupError(MultiError):
    pass



import asyncio

import aiotools


async def mytick(interval):
    print("tick")


async def run():
    t = aiotools.create_timer(mytick, 1.0)
    await asyncio.sleep(4)
    t.cancel()
    await t


if __name__ == "__main__":
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        loop.run_until_complete(run())
    finally:
        loop.stop()



from __future__ import annotations

import asyncio
import contextvars
from asyncio import events, exceptions, tasks
from contextvars import Context
from types import TracebackType
from typing import (
    Any,
    Self,
    TypeGuard,
    TypeVar,
)

from .taskcontext import ErrorCallback, LoopExceptionHandler, TaskContext
from .types import CoroutineLike

T = TypeVar("T")

__all__ = ("TaskScope",)


class TaskScope(TaskContext):
    """
    TaskScope is an asynchronous context manager which implements structured
    concurrency, i.e., "scoped" cancellation, over a set of child tasks.
    It terminates when all child tasks make conclusion (either results or exceptions).

    :class:`TaskScope` subclasses :class:`~aiotools.taskcontext.TaskContext`, but it
    mandates use of ``async with`` blocks to clarify which task is the parent of the
    child tasks spawned via :meth:`create_task()`.

    The key difference to :class:`asyncio.TaskGroup` is that it allows
    customization of the exception handling logic for unhandled child
    task exceptions, instead cancelling all pending child tasks upon any
    unhandled child task exceptions and collecting them as an :exc:`ExceptionGroup`.

    Since :class:`TaskScope` may be used for a long-running server context, unhandled
    child exceptions are NOT stored at all, but passed to the exception handler
    directly and immediately.  If you want to collect results and exceptions, please
    use :func:`~aiotools.utils.as_completed_safe()` or
    :func:`~aiotools.utils.gather_safe()`.

    Refer :class:`TaskContext` for the descriptions about the constructor arguments.

    Based on this customizability, :class:`~aiotools.supervisor.Supervisor` is a mere
    alias of :class:`TaskScope` with ``exception_handler=None``.

    .. versionadded:: 2.0
    """

    _tasks: set[asyncio.Task[Any]]
    _on_completed_fut: asyncio.Future[Any] | None
    _base_error: BaseException | None
    _entered: bool
    _exiting: bool
    _aborting: bool

    def __init__(
        self,
        *,
        exception_handler: ErrorCallback
        | LoopExceptionHandler
        | None = LoopExceptionHandler.TOKEN,
        context: contextvars.Context | None = None,
    ) -> None:
        super().__init__(exception_handler=exception_handler, context=context)
        # status flags
        self._entered = False
        self._exiting = False
        self._aborting = False
        # taskscope-specifics
        self._base_error = None
        self._on_completed_fut = None

    async def __aenter__(self) -> Self:
        if self._entered:
            raise RuntimeError(
                f"{type(self).__name__} {self!r} has been already entered"
            )
        self._entered = True

        if self._loop is None:
            self._loop = events.get_running_loop()

        self._parent_task = tasks.current_task(self._loop)
        if self._parent_task is None:
            raise RuntimeError(
                f"{type(self).__name__} {self!r} cannot determine the parent task"
            )

        return self

    async def __aexit__(
        self,
        et: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> bool | None:
        assert self._loop is not None
        assert self._parent_task is not None
        self._exiting = True

        if exc is not None and self._is_base_error(exc) and self._base_error is None:
            self._base_error = exc

        propagate_cancellation_error = exc if et is exceptions.CancelledError else None

        if et is not None:
            if not self._aborting:
                # Our parent task is being cancelled:
                #
                #    async with TaskGroup() as g:
                #        g.create_task(...)
                #        await ...  # <- CancelledError
                #
                # or there's an exception in "async with":
                #
                #    async with TaskGroup() as g:
                #        g.create_task(...)
                #        1 / 0
                #
                self.abort()

        prop_ex = await self._wait_completion()
        assert not self._tasks
        if prop_ex is not None:
            propagate_cancellation_error = prop_ex
        self._exited = True

        if self._base_error is not None:
            raise self._base_error

        # Propagate CancelledError as the child exceptions are handled separately.
        if propagate_cancellation_error:
            raise propagate_cancellation_error

        return None

    async def _wait_completion(self) -> BaseException | None:
        # We use while-loop here because "self._on_completed_fut"
        # can be cancelled multiple times if our parent task
        # is being cancelled repeatedly (or even once, when
        # our own cancellation is already in progress)
        assert self._loop is not None
        propagate_cancellation_error = None
        while self._tasks:
            if self._on_completed_fut is None:
                self._on_completed_fut = self._loop.create_future()
            try:
                await self._on_completed_fut
            except exceptions.CancelledError as ex:
                if not self._aborting:
                    # Our parent task is being cancelled:
                    #
                    #    async def wrapper():
                    #        async with TaskGroup() as g:
                    #            g.create_task(foo)
                    #
                    # "wrapper" is being cancelled while "foo" is
                    # still running.
                    propagate_cancellation_error = ex
                    self.abort(msg=ex.args[0] if ex.args else None)
            self._on_completed_fut = None
        return propagate_cancellation_error

    async def aclose(self) -> None:
        """
        Triggers cancellation and waits for completion.
        """
        self.abort(r"{self!r} is closed")
        await self._wait_completion()

    def create_task(
        self,
        coro: CoroutineLike[T],
        *,
        name: str | None = None,
        context: Context | None = None,
    ) -> tasks.Task[T]:
        """
        Create a new task in this scope and return it.
        Similar to :func:`asyncio.create_task()`.
        """
        if not self._entered:
            raise RuntimeError(f"{type(self).__name__} {self!r} has not been entered")
        if self._exiting and not self._tasks:
            raise RuntimeError(f"{type(self).__name__} {self!r} is finished")
        return self._create_task(coro, name=name, context=context)

    # Since Python 3.8 Tasks propagate all exceptions correctly,
    # except for KeyboardInterrupt and SystemExit which are
    # still considered special.

    def _is_base_error(self, exc: BaseException) -> TypeGuard[BaseException]:
        assert isinstance(exc, BaseException)
        return isinstance(exc, (SystemExit, KeyboardInterrupt))

    def abort(self, msg: str | None = None) -> None:
        """
        Triggers cancellation and immediately returns *without* waiting for completion.
        """
        self._aborting = True
        for t in self._tasks:
            if not t.done():
                t.cancel(msg=msg)

    def _on_task_done(self, task: asyncio.Task[Any]) -> None:
        assert self._loop is not None
        assert self._parent_task is not None
        self._tasks.discard(task)
        if self._on_completed_fut is not None and not self._tasks:
            if not self._on_completed_fut.done():
                self._on_completed_fut.set_result(True)
        if task.cancelled():
            return
        exc = task.exception()
        if exc is None:
            return

        # Instead of adding to ExceptionGroup, call the configured exception handler.
        self._handle_task_exception(task)

        is_base_error = self._is_base_error(exc)
        if is_base_error and self._base_error is None:
            self._base_error = exc

        if self._parent_task.done():
            # Not sure if this case is possible, but we want to handle
            # it anyways.
            self._loop.call_exception_handler({
                "message": (
                    f"Task {task!r} has errored out but its parent "
                    f"task {self._parent_task} is already completed"
                ),
                "exception": exc,
                "task": task,
            })
            return

        # If parent task *is not* being cancelled, we should just keep
        # running, unlike TaskGroup:
        #
        #    async def foo():
        #        try:
        #            async with TaskScope() as ts:
        #                ts.create_task(crash_soon())
        #                await something  # <- this should keep running
        #        except Exception:
        #            # Ignore any exceptions raised in the TaskScope
        #            pass
        #        await something_else     # <- unaffected as well



from __future__ import annotations

import asyncio
import contextvars
import enum
import warnings
from collections.abc import Callable
from contextvars import Context
from typing import (
    Any,
    TypeAlias,
    TypedDict,
    TypeVar,
)

from .cancel import cancel_and_wait
from .types import CoroutineLike

__all__ = (
    "ErrorArg",
    "ErrorCallback",
    "TaskContext",
)


class LoopExceptionHandler(enum.Enum):
    TOKEN = 0


class ErrorArg(TypedDict):
    # Intentionally designed as a typed dict to match with
    # the signature of asyncio stdlib's error callback handler.
    message: str
    exception: BaseException
    task: asyncio.Task[Any]


ErrorCallback: TypeAlias = Callable[[ErrorArg], None]
T = TypeVar("T")


class TaskContext:
    """
    TaskContext keeps the references to the child tasks during its lifetime,
    so that they can be terminated safely when shutdown is explicitly requested.

    This is the loosest form of child task managers among TaskScope, TaskGroup, and
    Supervisor, as it does not enforce structured concurrency but just provides a
    reference set to child tasks.

    You may replace existing patterns using :class:`weakref.WeakSet` to keep track
    of child tasks for a long-running server application with TaskContext.

    If ``exception_handler`` is not set (the default behavior), it will run
    :meth:`loop.call_exception_handler() <asyncio.loop.call_exception_handler>`
    with the context argument consisting of the ``message``, ``task`` (the child task
    that raised the exception), and ``exception`` (the exception object) fields.

    If it is set *None*, it will silently ignore the exception.

    If it is set as a callable function, it will invoke the specified callback
    function using the context argument same to that used when calling
    :meth:`loop.call_exception_handler() <asyncio.loop.call_exception_handler>`.

    If you provide ``context``, it will be passed to :meth:`create_task()` by default.

    .. versionadded:: 2.0
    """

    # We need to keep track of strong references as the event loop keeps weak references to
    # fire-and-forget tasks until Python 3.13. In Python 3.14, it will be changed as
    # per-thread doubly-linked lists to support free-threaded (nogil) setups to avoid
    # lock contention in the weakref subsystem.
    _tasks: set[asyncio.Task[Any]]
    _parent_task: asyncio.Task[Any] | None
    _loop: asyncio.AbstractEventLoop | None
    _exception_handler: ErrorCallback | LoopExceptionHandler | None
    _default_context: contextvars.Context | None
    _entered: bool
    _exited: bool
    _aborting: bool

    def __init__(
        self,
        *,
        exception_handler: ErrorCallback
        | LoopExceptionHandler
        | None = LoopExceptionHandler.TOKEN,
        context: contextvars.Context | None = None,
    ) -> None:
        self._loop = None
        self._tasks = set()
        self._parent_task = None
        self._exception_handler = exception_handler
        self._default_context = context
        # status flags
        self._entered = False
        self._exited = False
        self._aborting = False

    def __del__(self) -> None:
        loc: str = "<loc>"  # TODO: implement
        if self._entered and not self._exited:
            warnings.warn(
                f"TaskContext initialized at {loc} is not properly "
                "terminated until it is garbage-collected.",
                category=ResourceWarning,
            )

    def __repr__(self) -> str:
        info = [""]
        if self._tasks:
            info.append(f"tasks={len(self._tasks)}")
        if self._aborting:
            info.append("cancelling")
        elif self._entered:
            info.append("entered")

        info_str = " ".join(info)
        return f"<{type(self).__name__} {info_str}>"

    async def aclose(self) -> None:
        """
        Triggers cancellation and waits for completion.
        """
        self._aborting = True
        try:
            # NOTE: unhandled exceptions are captured by our own task-done handler.
            await asyncio.gather(
                *(cancel_and_wait(t) for t in self._tasks),
                return_exceptions=True,
            )
        finally:
            self._exited = True

    def create_task(
        self,
        coro: CoroutineLike[T],
        *,
        name: str | None = None,
        context: Context | None = None,
    ) -> asyncio.Task[T]:
        """
        Create a new task in this scope and return it.
        Similar to :func:`asyncio.create_task()`.
        """
        if not self._entered:
            self._entered = True
            self._loop = asyncio.get_running_loop()
        parent_task = asyncio.current_task()
        assert parent_task is not None
        if self._parent_task is not None and parent_task is not self._parent_task:
            raise RuntimeError(
                "{type(self).__name__} must be used within a single parent task."
            )
        self._parent_task = parent_task
        return self._create_task(coro, name=name, context=context)

    def _create_task(
        self,
        coro: CoroutineLike[T],
        *,
        name: str | None = None,
        context: Context | None = None,
    ) -> asyncio.Task[T]:
        assert self._loop is not None
        if self._aborting:
            raise RuntimeError(f"{type(self).__name__} {self!r} is shutting down")
        task: asyncio.Task[T] = self._loop.create_task(
            coro,
            name=name,
            context=self._default_context if context is None else context,
        )
        # optimization: Immediately call the done callback if the task is
        # already done (e.g. if the coro was able to complete eagerly),
        # and skip scheduling a done callback
        if task.done():
            self._on_task_done(task)
        else:
            self._tasks.add(task)
            task.add_done_callback(self._on_task_done)
        return task

    def _on_task_done(self, task: asyncio.Task[Any]) -> None:
        self._tasks.discard(task)
        if task.cancelled():
            return
        exc = task.exception()
        if exc is None:
            return
        self._handle_task_exception(task)

    def _handle_task_exception(self, task: asyncio.Task[Any]) -> None:
        assert self._loop is not None
        exc = task.exception()
        assert exc is not None
        match self._exception_handler:
            case None:
                pass  # deliberately set to ignore errors
            case func if callable(func):
                func({
                    "message": (
                        f"Task {task!r} has errored inside the parent "
                        f"task {self._parent_task}"
                    ),
                    "exception": exc,
                    "task": task,
                })
            case LoopExceptionHandler():
                self._loop.call_exception_handler({
                    "message": (
                        f"Task {task!r} has errored inside the parent "
                        f"task {self._parent_task}"
                    ),
                    "exception": exc,
                    "task": task,
                })
            case _:
                raise RuntimeError(
                    f"Invalid error handler: {self._exception_handler!r}"
                )



from __future__ import annotations

import asyncio
from typing import Any

__all__ = ("cancel_and_wait",)


async def cancel_and_wait(
    task: asyncio.Task[Any],
    /,
    msg: str | None = None,
) -> None:
    """
    Safely cancels and waits until the given task is concluded as cancelled.
    If the task is already terminated, it does nothing.

    When the caller of this function is cancelled during waiting,
    cancellation is transparently raised up.  Otherwise, it consumes and absorbs
    :exc:`asyncio.CancelledError` raised while awaiting the cancelled task.
    This is the key point of this function: it guarantees the structural safety by
    ensuring the caller's control over cancellation.

    If the task does not re-raise :exc:`asyncio.CancelledError` when awaited,
    it explicitly raises :exc:`asyncio.InvalidStateError` to warn that the task
    should transparently raise up injected cancellations.

    See the discussion in https://github.com/python/cpython/issues/103486 for details.

    .. note::

       If you use :ref:`eager task factories <python:eager-task-factory>` in Python
       3.12 or later, tasks may have completed already if they do not contain any
       context switches (awaits) even when they are cancelled immediately after
       :func:`asyncio.create_task()` returns.
       In such cases, ``cancel_and_wait()`` becomes a no-op.

    .. versionadded:: 2.0
    """
    if task.done():
        return
    # After our cancellation, cancelling() should be incremented by 1.
    # If its incremented by more than 1, it means cancel was requested externally.
    # In that case CancelledError should be raised to also end waiting task.
    cancelling_expected = task.cancelling() + 1
    task.cancel(msg)
    try:
        await task
    except asyncio.CancelledError:
        if task.cancelling() != cancelling_expected:
            raise
        else:
            return  # this is the only non-exceptional return
    else:
        raise asyncio.InvalidStateError(
            f"The cancelled task {task!r} did not raise up cancellation."
        )
